"""
Atmosphere service driver.

Driver classes define interfaces and implement functionality using providers.
"""
from abc import ABCMeta, abstractmethod
import copy
from datetime import datetime
import sys
import time

from libcloud.compute.deployment import ScriptDeployment
from libcloud.compute.deployment import MultiStepDeployment
from libcloud.compute.types import DeploymentError

from threepio import logger

from rtwo import settings
from rtwo.drivers.common import LoggedScriptDeployment

from rtwo.exceptions import MissingArgsException, ServiceException

from rtwo.models.provider import AWSProvider
from rtwo.models.provider import EucaProvider
from rtwo.models.provider import OSProvider

from rtwo.models.identity import AWSIdentity
from rtwo.models.identity import EucaIdentity
from rtwo.models.identity import OSIdentity

from rtwo.mixins.driver import APIFilterMixin, MetaMixin,\
    InstanceActionMixin


class BaseDriver():
    """
    BaseDriver lists a basic set of expected functionality for an esh-driver.
    Abstract class - Should not be instantiated!!
    """

    __metaclass__ = ABCMeta

    _connection = None

    identity = None

    provider = None

    identityCls = None

    providerCls = None

    @abstractmethod
    def __init__(self, identity, provider):
        raise NotImplementedError

    @abstractmethod
    def list_instances(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def list_machines(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def list_sizes(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def list_locations(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def create_instance(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def deploy_instance(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def reset_network(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def reboot_instance(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def destroy_instance(self, *args, **kwargs):
        raise NotImplementedError

    def start_instance(self, *args, **kwargs):
        raise NotImplementedError

    def stop_instance(self, *args, **kwargs):
        raise NotImplementedError

    def resume_instance(self, *args, **kwargs):
        raise NotImplementedError

    def suspend_instance(self, *args, **kwargs):
        raise NotImplementedError

    def resize_instance(self, *args, **kwargs):
        raise NotImplementedError


class VolumeDriver():
    """
    VolumeDriver provides basic storage volume functionality for libcloud
    or esh-drivers.
    Abstract class - Should not be instantiated!!
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def list_volumes(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def create_volume(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def destroy_volume(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def boot_volume(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def attach_volume(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def detach_volume(self, *args, **kwargs):
        raise NotImplementedError


class LibcloudDriver(BaseDriver, VolumeDriver, APIFilterMixin):
    """
    Provides direct access to the libcloud methods and data.
    """

    def __init__(self, provider, identity, **provider_credentials):
        if provider is None or identity is None:
            raise MissingArgsException(
                'LibcloudDriver is Missing Required Identity and/or Provider.')
        self.identity = identity
        self.provider = provider
        self._connection = self.provider.get_driver(
                                            self.identity,
                                            **provider_credentials)

    def list_instances(self, *args, **kwargs):
        return self._connection.list_nodes()

    def list_machines(self, *args, **kwargs):
        logger.debug("Call made to list_machines: %s - %s"
                     % (self.provider.identifier,
                     self.identity.credentials['key']))
        return self._connection.list_images()

    def list_sizes(self, *args, **kwargs):
        logger.debug("Call made to list_sizes: %s - %s"
                     % (self.provider.identifier,
                     self.identity.credentials['key']))
        return self._connection.list_sizes()

    def list_locations(self, *args, **kwargs):
        return self._connection.list_locations()

    def create_instance(self, *args, **kwargs):
        return self._connection.create_node(*args, **kwargs)

    def deploy_instance(self, *args, **kwargs):
        return self._connection.deploy_node(*args, **kwargs)

    def reset_network(self, *args, **kwargs):
        return self._connection.reset_network(*args, **kwargs)

    def reboot_instance(self, *args, **kwargs):
        return self._connection.reboot_node(*args, **kwargs)

    def destroy_instance(self, *args, **kwargs):
        return self._connection.destroy_node(*args, **kwargs)

    def list_volumes(self, *args, **kwargs):
        return self._connection.list_volumes(*args, **kwargs)

    def create_volume(self, *args, **kwargs):
        return self._connection.create_volume(*args, **kwargs)

    def destroy_volume(self, *args, **kwargs):
        return self._connection.destroy_volume(*args, **kwargs)

    def attach_volume(self, *args, **kwargs):
        return self._connection.attach_volume(*args, **kwargs)

    def detach_volume(self, *args, **kwargs):
        return self._connection.detach_volume(*args, **kwargs)


class EshDriver(LibcloudDriver, MetaMixin):
    """
    """

    @classmethod
    def settings_init(cls):
        raise ServiceException('Settings init not available for this class')

    def __init__(self, provider, identity, **provider_credentials):
        super(EshDriver, self).__init__(provider, identity, **provider_credentials)
        if not(isinstance(provider, self.providerCls)
           and isinstance(identity, self.identityCls)):
            raise ServiceException('Wrong Provider or Identity')

    def list_all_volumes(self, *args, **kwargs):
        """
        Return the InstanceClass representation of a libcloud node
        """
        return self.provider.volumeCls.get_volumes(
            self._connection.ex_list_all_volumes())

    def list_all_instances(self, **kwargs):
        """
        Return the InstanceClass representation of a libcloud node
        """
        return self.provider.instanceCls.get_instances(
            self._connection.ex_list_all_instances(), self.provider)

    def get_instance(self, instance_id, *args, **kwargs):
        """
        Return the InstanceClass representation of a libcloud node
        """
        instances = self.list_all_instances()
        instance = [inst for inst in instances if inst.id == instance_id]
        if not instance:
            return None
        return instance[0]

    def list_instances(self, **kwargs):
        """
        Return the InstanceClass representation of a libcloud node
        """
        return self.provider.instanceCls.get_instances(
            super(EshDriver, self).list_instances(), self.provider)

    def list_machines(self, *args, **kwargs):
        """
        Return the MachineClass representation of a libcloud NodeImage
        """
        return self.provider.machineCls.get_cached_machines(
            self.provider.identifier,
            super(EshDriver, self).list_machines, *args, **kwargs)

    def list_sizes(self, *args, **kwargs):
        """
        Return the SizeClass representation of a libcloud NodeSize
        """
        return self.provider.sizeCls.get_sizes(
            self.provider,
            super(EshDriver, self).list_sizes)

    def list_locations(self, *args, **kwargs):
        return super(EshDriver, self).list_locations()

    def create_instance(self, *args, **kwargs):
        """
        Return the InstanceClass representation of a libcloud node
        """
        logger.debug(str(args))
        logger.debug(str(kwargs))
        return self.provider.instanceCls(
            super(EshDriver, self).create_instance(*args, **kwargs),
            self.provider)

    def deploy_instance(self, *args, **kwargs):
        return self.provider.instanceCls(
            super(EshDriver, self).deploy_instance(*args, **kwargs),
            self.provider)

    def reset_network(self, *args, **kwargs):
        return super(EshDriver, self).reset_network(*args, **kwargs)

    def reboot_instance(self, *args, **kwargs):
        return super(EshDriver, self).reboot_instance(*args, **kwargs)

    def start_instance(self, *args, **kwargs):
        return super(EshDriver, self).start_instance(*args, **kwargs)

    def stop_instance(self, *args, **kwargs):
        return super(EshDriver, self).stop_instance(*args, **kwargs)

    def resume_instance(self, *args, **kwargs):
        return super(EshDriver, self).resume_instance(*args, **kwargs)

    def suspend_instance(self, *args, **kwargs):
        return super(EshDriver, self).suspend_instance(*args, **kwargs)

    def destroy_instance(self, *args, **kwargs):
        return super(EshDriver, self).destroy_instance(*args, **kwargs)

    def boot_volume(self, *args, **kwargs):
        return self.provider.instanceCls(
                super(EshDriver, self).boot_volume(*args, **kwargs),
                self.provider)

    def list_volumes(self, *args, **kwargs):
        return self.provider.volumeCls.get_volumes(
            super(EshDriver, self).list_volumes(*args, **kwargs))

    def create_volume(self, *args, **kwargs):
        return super(EshDriver, self).create_volume(*args, **kwargs)

    def destroy_volume(self, *args, **kwargs):
        return super(EshDriver, self).destroy_volume(*args, **kwargs)

    def attach_volume(self, *args, **kwargs):
        return super(EshDriver, self).attach_volume(*args, **kwargs)

    def detach_volume(self, *args, **kwargs):
        return super(EshDriver, self).detach_volume(*args, **kwargs)


class OSDriver(EshDriver, InstanceActionMixin):
    """
    """
    providerCls = OSProvider

    identityCls = OSIdentity

    @classmethod
    def admin_init(cls, id_credentials, prov_credentials):
        """
        To be used instead of settings_init, Expects the following keys for
        each credentials dict:
        id_credentials:
        * Username
        * Password
        * (Tenant/Project) Name
        Prov_credentials:
        * region_name
        * auth_url
        """
        try:
            username = id_credentials.pop('username')
            password = id_credentials.pop('password')
            tenant_name = id_credentials.pop('tenant_name')
        except:
            raise ServiceException(
                'Settings init not available for this class:'
                'Expected settings.OPENSTACK_ARGS with'
                'username/password/tenant_name fields')
        OSProvider.set_meta()
        provider = OSProvider()
        identity = OSIdentity(provider, username, password,
                              ex_tenant_name=tenant_name,
                              **prov_credentials)
        driver = cls(provider, identity)
        return driver

    @classmethod
    def settings_init(cls):
        """
        DEPRECATED:
        """
        os_args = {}
        try:
            username = os_args.pop('username')
            password = os_args.pop('password')
            tenant_name = os_args.pop('tenant_name')
        except:
            raise ServiceException(
                'Settings init not available for this class:'
                'Expected settings.OPENSTACK_ARGS with'
                'username/password/tenant_name fields')
        OSProvider.set_meta()
        provider = OSProvider()
        identity = OSIdentity(provider, username, password,
                              ex_tenant_name=tenant_name)
        driver = cls(provider, identity)
        return driver

    def __init__(self, provider, identity, **provider_credentials):
        super(OSDriver, self).__init__(provider, identity, **provider_credentials)
        #Set connection && force_service_region
        self._connection._ex_force_service_region =\
        self._connection.connection.service_region =\
            provider.options.get('region_name')

    def list_machines(self, *args, **kwargs):
        """
        This openstack specific implementation caches machine lists
        using tenant_name as the identifier

        Return the MachineClass representation of a libcloud NodeImage
        """
        #TODO: I don't like this implementation.. -Steve
        # Currently there is no 'override' if i dont want to use ex-tenant-name
        identifier = self.identity.credentials.get('ex_tenant_name')
        if not identifier:
            logger.debug("Could not find tenant_name, "
                         "falling back to provider identifier")
            identifier = self.provider.identitifier
        return self.provider.machineCls.get_cached_machines(
            identifier,
            super(EshDriver,self).list_machines, *args, **kwargs)


    def deploy_init_to(self, *args, **kwargs):
        """
        Creates a multi script deployment to prepare and call
        the latest init script
        TODO: Add versioning for 30+
        """
        if args:
            instance = args[0]
        else:
            raise MissingArgsException("Missing instance argument.")
        if isinstance(self.identity.user, basestring):
            username = self.identity.user
        else:
            # Django.contrib.auth.models.User
            username = self.identity.user.username
        atmo_init = "/usr/sbin/atmo_init_full.py"
        server_atmo_init = "/init_files/30/atmo_init_full.py"
        script_init = ScriptDeployment(
            'if [ ! -d "/var/log/atmo" ];then\n'
            'mkdir -p /var/log/atmo\n'
            'fi\n'
            'if [ ! -f "/var/log/atmo/deploy.log" ]; then\n'
            'touch /var/log/atmo/deploy.log\n'
            'fi',
            name="./deploy_init_log.sh")
        #These requirements are for Editors, Shell-in-a-box
        do_ubuntu = "apt-get update;apt-get install -y emacs vim wget "\
                    + "language-pack-en make gcc g++ gettext texinfo "\
                    + "autoconf automake"
        do_centos = "yum install -y emacs vim-enhanced wget make "\
                    + "gcc gettext texinfo autoconf automake python-simplejson"
        script_deps = LoggedScriptDeployment(
            "distro_cat=`cat /etc/*-release`\n"
            + "if [[ $distro_cat == *Ubuntu* ]]; then\n"
            + do_ubuntu
            + "\nelse if [[ $distro_cat == *CentOS* ]];then\n"
            + do_centos
            + "\nfi\nfi",
            name="./deploy_deps.sh",
            logfile="/var/log/atmo/deploy.log")
        script_wget = LoggedScriptDeployment(
            "wget -O %s %s%s" % (atmo_init, settings.SERVER_URL,
                                 server_atmo_init),
            name='./deploy_wget_atmoinit.sh',
            logfile='/var/log/atmo/deploy.log')
        script_chmod = LoggedScriptDeployment(
            "chmod a+x %s" % atmo_init,
            name='./deploy_chmod_atmoinit.sh',
            logfile='/var/log/atmo/deploy.log')
        instance_token = kwargs.get('token', '')
        if not instance_token:
            instance_token = instance.id
        awesome_atmo_call = "%s --service_type=%s --service_url=%s"
        awesome_atmo_call += " --server=%s --user_id=%s"
        awesome_atmo_call += " --token=%s --name=\"%s\""
        awesome_atmo_call += " --vnc_license=%s"
        awesome_atmo_call %= (
            atmo_init,
            "instance_service_v1",
            settings.INSTANCE_SERVICE_URL,
            settings.SERVER_URL,
            username,
            instance_token,
            instance.name,
            settings.ATMOSPHERE_VNC_LICENSE)
        #kludge: weirdness without the str cast...
        str_awesome_atmo_call = str(awesome_atmo_call)
        #logger.debug(isinstance(str_awesome_atmo_call, basestring))
        script_atmo_init = LoggedScriptDeployment(
            str_awesome_atmo_call,
            name='./deploy_call_atmoinit.sh',
            logfile='/var/log/atmo/deploy.log')
        script_rm_scripts = LoggedScriptDeployment(
            "rm -rf ~/deploy_*",
            name='./deploy_remove_scripts.sh',
            logfile='/var/log/atmo/deploy.log')
        msd = MultiStepDeployment([script_init,
                                   script_deps,
                                   script_wget,
                                   script_chmod,
                                   script_atmo_init,
                                   script_rm_scripts])
        kwargs.update({'deploy': msd})

        private_key = "/opt/dev/atmosphere/extras/ssh/id_rsa"
        kwargs.update({'ssh_key': private_key})

        kwargs.update({'timeout': 120})

        return self.deploy_to(instance, *args, **kwargs)

    def deploy_to(self, *args, **kwargs):
        """
        Deploy to an instance.
        """
        if args:
            instance = args[0]
        else:
            raise MissingArgsException("Missing instance argument.")
        if not kwargs.get('deploy'):
            raise MissingArgsException("Missing deploy argument.")
        if not kwargs.get('ssh_key'):
            private_key = "/opt/dev/atmosphere/extras/ssh/id_rsa"
            kwargs.update({'ssh_key': private_key})
        if not kwargs.get('timeout'):
            kwargs.update({'timeout': 120})

        #Scrub deploy steps if they exist.
        deploy = kwargs.get('deploy')
        if hasattr(deploy, "steps"):
            deploy.steps = [s for s in deploy.steps if s is not None]

        logger.info("Attempting deployment to node")
        node = instance
        #Get the libcloud node, not the eshInstance
        if hasattr(instance, '_node'):
            node = instance._node
        self._connection.ex_deploy_to_node(node,
                                           *args, **kwargs)
        return True

    def deploy_instance(self, *args, **kwargs):
        """
        Deploy instance.

        NOTE: This is blocking and uses the blocking create_node.
        """
        if not kwargs.get('deploy'):
            raise MissingArgsException("Missing deploy argument.")
        username = self.identity.user.username

        private_key = "/opt/dev/atmosphere/extras/ssh/id_rsa"
        kwargs.update({'ssh_key': private_key})
        kwargs.update({'timeout': 120})

        #cloudinit_script = prepare_cloudinit_script()
        #kwargs.update({'ex_userdata': cloudinit_script})
        try:
            self.deploy_node(*args, **kwargs)
        except DeploymentError as de:
            logger.error(sys.exc_info())
            logger.error(de.value)
            return False
        return True

    def destroy_instance(self, *args, **kwargs):
        return self._connection.destroy_node(*args, **kwargs)

    def start_instance(self, *args, **kwargs):
        return self._connection.ex_start_node(*args, **kwargs)

    def stop_instance(self, *args, **kwargs):
        return self._connection.ex_stop_node(*args, **kwargs)

    def suspend_instance(self, *args, **kwargs):
        return self._connection.ex_suspend_node(*args, **kwargs)

    def resume_instance(self, *args, **kwargs):
        return self._connection.ex_resume_node(*args, **kwargs)

    def resize_instance(self, *args, **kwargs):
        return self._connection.ex_resize(*args, **kwargs)

    def reset_network(self, *args, **kwargs):
        return self._connection.reset_network(*args, **kwargs)

    def reboot_instance(self, *args, **kwargs):
        return self._connection.reboot_node(*args, **kwargs)

    def boot_volume(self, *args, **kwargs):
        return self._connection.ex_boot_volume(*args, **kwargs)

    def confirm_resize_instance(self, *args, **kwargs):
        return self._connection.ex_confirm_resize(*args, **kwargs)

    def revert_resize_instance(self, *args, **kwargs):
        return self._connection.ex_revert_resize(*args, **kwargs)

    def _add_floating_ip(self, instance, *args, **kwargs):
        return self._connection.neutron_associate_ip(instance, *args, **kwargs)

    def _del_floating_ip(self, instance, *args, **kwargs):
        return self._connection.neutron_disassociate_ip(instance, *args, **kwargs)

    def _clean_floating_ip(self, *args, **kwargs):
        return self._connection.ex_clean_floating_ip(**kwargs)

    def _is_inactive_instance(self, instance):
        #Other things may need to be tested
        status = instance.extra['status']
        task = instance.extra['task']
        power = instance.extra['power']
        if status in ['paused',
                'suspended','stopped','shutoff',]:
            return True
        if task in ['suspending','shutting-off',
                    ]:
            return True
        return False

    def _is_active_instance(self, instance):
        #Other things may need to be tested
        status = instance.extra['status']
        task = instance.extra['task']
        power = instance.extra['power']
        if status in ['active','build','resize']:
            if task in ['deleting', 'suspending']:
                result=False
            result=True
        elif task in ['resuming', 'powering-on',
                    'verify-resize', 'resize_reverting', 'resize_confirming']:
            result=True
        else:
            result=False
        logger.info("Instance: %s Status: %s-%s Active:%s" % (instance.id,
            status, task, result))
        return result


class AWSDriver(EshDriver):
    """
    """
    providerCls = AWSProvider

    identityCls = AWSIdentity

    def deploy_instance(self, *args, **kwargs):
        """
        Deploy an AWS node.
        """
        username = self.identity.user.username
        atmo_init = "/usr/sbin/atmo_init_full.py"
        server_atmo_init = "/init_files/30/atmo_init_full.py"
        script_deps = ScriptDeployment(
            "sudo apt-get install -y emacs vim wget")
        script_wget = ScriptDeployment(
            "sudo wget -O %s %s%s" %
            (atmo_init, settings.SERVER_URL, server_atmo_init))
        script_chmod = ScriptDeployment("sudo chmod a+x %s" % atmo_init)
        instance_token = kwargs.get('token', '')
        awesome_atmo_call = "sudo %s --service_type=%s --service_url=%s"
        awesome_atmo_call += " --server=%s --user_id=%s --token=%s"
        awesome_atmo_call += " --name=%s &> %s"
        awesome_atmo_call %= (
            atmo_init,
            "instance_service_v1",
            settings.INSTANCE_SERVICE_URL,
            settings.SERVER_URL,
            username,
            instance_token,
            kwargs.get('name', ''),
            '/var/log/atmo_init_full.err')
        logger.debug(awesome_atmo_call)
        str_awesome_atmo_call = str(awesome_atmo_call)
        #kludge: weirdness without the str cast...
        script_atmo_init = ScriptDeployment(str_awesome_atmo_call)
        private_key = ("/opt/dev/atmosphere/extras/ssh/id_rsa")
        scripts = [script_deps,
                   script_wget,
                   script_chmod,
                   script_atmo_init]
        for s in scripts:
            logger.debug(s.name)
            s.name = s.name.replace('/root', '/home/ubuntu')
            logger.debug(s.name)
        msd = MultiStepDeployment(scripts)
        kwargs.update({'ex_keyname': 'dalloway-key'})
        kwargs.update({'ssh_username': 'ubuntu'})
        kwargs.update({'ssh_key': private_key})
        kwargs.update({'deploy': msd})
        kwargs.update({'timeout': 400})

        instance = super(AWSDriver, self).deploy_instance(*args, **kwargs)
        created = datetime.strptime(instance.extra['created'],
                                    "%Y-%m-%dT%H:%M:%SZ")
        # NOTE: Removed for rtwo port. Moved to service tasks.
        # send_instance_email(username, instance.id, instance.ip,
        # created, username)
        return instance


    def create_volume(self, *args, **kwargs):
        if 'description' in kwargs:
            kwargs.pop('description')
        return super(EshDriver, self).create_volume(*args, **kwargs)


class EucaDriver(EshDriver):
    """
    """
    providerCls = EucaProvider

    identityCls = EucaIdentity

    def deploy_instance(self, *args, **kwargs):
        raise NotImplementedError

    def resume_instance(self, *args, **kwargs):
        raise NotImplementedError

    def suspend_instance(self, *args, **kwargs):
        raise NotImplementedError

    def start_instance(self, *args, **kwargs):
        raise NotImplementedError

    def stop_instance(self, *args, **kwargs):
        raise NotImplementedError



"""
Atmosphere service provider.

"""
from abc import ABCMeta, abstractmethod

from libcloud.compute.types import Provider as LProvider

from threepio import logger

from rtwo.drivers.openstack_facade import OpenStack_Esh_NodeDriver
from rtwo.drivers.eucalyptus import Eucalyptus_Esh_NodeDriver
from rtwo.drivers.aws import Esh_EC2NodeDriver

from rtwo.exceptions import ServiceException


def lc_provider_id(provider):
    """
    Get the libcloud Provider using our service provider.

    Return the libcloud.compute Provider value.
    """
    p = None
    try:
        p = LProvider.__dict__[provider.location]
    except Exception as e:
        logger.warn("Unable to find provider location: %s." %
                    provider.location)
        raise ServiceException(e)
    return p


class BaseProvider(object):
    __metaclass__ = ABCMeta

    identity = None

    lc_driver = None

    options = {}

    #Used to support multi-cloud
    identifier = ''

    location = ''

    name = ''

    identityCls = None

    instanceCls = None

    machineCls = None

    sizeCls = None

    @abstractmethod
    def __init__(self, *args, **kwargs):
        raise NotImplemented

    @abstractmethod
    def provider_id(self):
        raise NotImplemented

    @abstractmethod
    def set_options(self):
        raise NotImplemented

    @abstractmethod
    def get_driver(self, *args, **kwargs):
        raise NotImplemented

    def __repr__(self):
        return '%s (%s)' % (self.__class__, self.identifier)


class Provider(BaseProvider):

    def __init__(self, url=None, identifier=None):
        """
        :param url: url to use for this provider.
        """
        if url:
            self.options.update(Provider.parse_url(kwargs['url']))
        if not identifier:
            identifier = self.location
        self.identifier = identifier

    def set_options(self, provider_credentials):
        """
        Get provider specific options.

        Return provider specific options in a dict.
        """
        self.options = {}
        self.options.update(provider_credentials)
        self.options.update(self.identity.credentials)
        return self.options

    @classmethod
    def parse_url(cls, url):
        """
        Parse the url into the options dictionary.
        """
        from urlparse import urlparse
        urlobj = urlparse(url)
        options = {}
        options['host'] = urlobj.hostname
        options['port'] = urlobj.port
        options['path'] = urlobj.path
        options['secure'] = urlobj.scheme == 'https'
        return options

    def provider_id(self):
        try:
            return lc_provider_id(self)
        except Exception as e:
            raise ServiceException(e)


class AWSProvider(Provider):

    name = 'Amazon EC2'

    location = 'EC2_US_EAST'

    @classmethod
    def set_meta(cls):
        from rtwo.models.identity import AWSIdentity
        from rtwo.models.machine import AWSMachine
        from rtwo.models.instance import AWSInstance
        from rtwo.models.size import AWSSize
        from rtwo.models.volume import AWSVolume
        from rtwo.meta import AWSMeta
        cls.identityCls = AWSIdentity
        cls.machineCls = AWSMachine
        cls.instanceCls = AWSInstance
        cls.sizeCls = AWSSize
        cls.volumeCls = AWSVolume
        cls.metaCls = AWSMeta

    def set_options(self, provider_credentials):
        """
        Get provider specific options.

        Return provider specific options in a dict.
        """
        self.options = {}
        self.options.update(provider_credentials)
        self.options.update(self.identity.credentials)
        return self.options

    def get_driver(self, identity, **provider_credentials):
        """
        Get the libcloud driver using our service identity.

        Return the libcloud.compute driver class.
        """
        self.identity = identity
        self.lc_driver = Esh_EC2NodeDriver
        self.set_options(provider_credentials)
        return self.lc_driver(key=self.options['key'],
                              secret=self.options['secret'])


class AWSUSWestProvider(AWSProvider):

    location = 'EC2_US_WEST'


class AWSUSEastProvider(AWSProvider):

    location = 'EC2_US_EAST'


class EucaProvider(Provider):

    name = 'Eucalyptus'

    # This need to be in all caps to match lib cloud.
    location = 'EUCALYPTUS'

    @classmethod
    def set_meta(cls):
        from rtwo.models.identity import EucaIdentity
        from rtwo.models.machine import EucaMachine
        from rtwo.models.instance import EucaInstance
        from rtwo.models.size import EucaSize
        from rtwo.models.volume import EucaVolume
        from rtwo.meta import EucaMeta
        cls.identityCls = EucaIdentity
        cls.machineCls = EucaMachine
        cls.instanceCls = EucaInstance
        cls.sizeCls = EucaSize
        cls.volumeCls = EucaVolume
        cls.metaCls = EucaMeta

    def set_options(self, provider_credentials):
        """
        Get provider specific credentials.

        Return any provider specific credentials in a dict.
        """
        self.options = {'host': '128.196.172.136',
                        'secure': False,
                        'port': 8773,
                        'path': '/services/Eucalyptus'}
        # Options default to provider
        self.options.update(provider_credentials)
        self.options.update(self.identity.credentials)
        return self.options

    def get_driver(self, identity, **provider_credentials):
        """
        Get the libcloud driver using our service identity.

        Return the libcloud.compute driver class.
        """
        self.lc_driver = Eucalyptus_Esh_NodeDriver

        self.identity = identity
        self.set_options(provider_credentials)

        return self.lc_driver(key=self.options['key'],
                              secret=self.options['secret'],
                              secure=self.options['secure'] != False,
                              host=self.options['host'],
                              port=self.options['port'],
                              path=self.options['path'])


class OSProvider(Provider):

    name = 'OpenStack'

    location = 'OPENSTACK'

    @classmethod
    def set_meta(cls):
        from rtwo.models.identity import OSIdentity
        from rtwo.models.machine import OSMachine
        from rtwo.models.instance import OSInstance
        from rtwo.models.size import OSSize
        from rtwo.models.volume import OSVolume
        from rtwo.meta import OSMeta
        cls.identityCls = OSIdentity
        cls.machineCls = OSMachine
        cls.instanceCls = OSInstance
        cls.sizeCls = OSSize
        cls.volumeCls = OSVolume
        cls.metaCls = OSMeta

    def set_options(self, provider_credentials):
        """
        Get provider specific options.

        Return provider specific options in a dict.
        """
        self.options = {'secure': 'False',
                        'ex_force_auth_version': '2.0_password',
                        'ex_domain_name': None,
                        'ex_force_auth_url': None,
                        'ex_force_base_url': None,
                        'ex_force_auth_token': None
                       }
        self.options.update(provider_credentials)
        self.options.update(self.identity.credentials)
        #Basic validation to avoid 'hard-to-reason-about-errors'
        if self.options['ex_force_auth_version'] == '2.0_password' and 'tokens' not in self.options['ex_force_auth_url']:
           raise ValueError("Conflicting options -- If you are using '2.0_password' for Keystone Authentication, your auth_url should include '/v2.0/tokens' in addition to the http(s)://hostname:port")
        elif self.options['ex_force_auth_version'] == '3.x_password' and 'tokens' in self.options['ex_force_auth_url']:
           raise ValueError("Conflicting options -- If you are using '3.x_password' for Keystone Authentication, your auth_url should NOT include '/v2.0/tokens'. Instead use: http(s)://hostname:port")

        return self.options

    def get_driver(self, identity, **provider_credentials):
        """
        Get the libcloud driver using our service identity.

        Return the libcloud.compute driver class.
        """
        self.identity = identity
        self.lc_driver = OpenStack_Esh_NodeDriver
        self.set_options(provider_credentials)
        return self.lc_driver(
            key=self.options['key'],
            secret=self.options['secret'],
            secure=self.options['secure'] != 'False',
            ex_domain_name=self.options['ex_domain_name'],
            ex_force_base_url=self.options['ex_force_base_url'],
            ex_force_auth_url=self.options['ex_force_auth_url'],
            ex_force_auth_version=self.options['ex_force_auth_version'],
            ex_force_auth_token=self.options['ex_force_auth_token'],
            ex_tenant_name=self.options['ex_tenant_name'])


class OSValhallaProvider(OSProvider):

    region_name = "ValhallaRegion"

    def set_options(self, provider_credentials):
        """
        """
        super(OSValhallaProvider, self).set_options(provider_credentials)


class OSMidgardProvider(OSProvider):

    region_name = "MidgardRegion"

    def set_options(self, provider_credentials):
        """
        """
        super(OSMidgardProvider, self).set_options(provider_credentials)



"""
Atmosphere service compute.
"""
from threepio import logger

from rtwo import settings

from rtwo.provider import AWSProvider, EucaProvider, OSProvider
from rtwo.driver import EucaDriver, AWSDriver

from libcloud.common.types import InvalidCredsError


OSProvider.set_meta()




"""
Extension of libcloud's Eucalyptus Node Driver.
"""
import re, copy, string
import base64, hmac, time
from datetime import datetime

from hashlib import sha256

from libcloud.utils.xml import findall, findtext, findattr
from libcloud.compute.base import NodeSize, StorageVolume, NodeImage
from libcloud.compute.types import Provider
from libcloud.compute.drivers.ec2 import EC2NodeDriver, NAMESPACE

from libcloud.utils.py3 import urlencode
from libcloud.utils.py3 import urlquote, b

from threepio import logger

class Esh_EC2NodeDriver(EC2NodeDriver):
    """
    Amazon EC2 Node Driver
    """

    def _get_attachment_set(self, element):
        """
        TEST
        """
        attachment_set = {}
        for elem in element:
            for key in ['volumeId', 'instanceId', 'device', 'status', 'attachTime']:
                value = findtext(element=elem, xpath=key, namespace=NAMESPACE)
                if value:
                    attachment_set[key] = value
        if attachment_set.get('attachTime', None):
            if '.' in attachment_set['attachTime']:
                attach_date = datetime.strptime(attachment_set['attachTime'], '%Y-%m-%dT%H:%M:%S.%fZ')
            else:
                attach_date = datetime.strptime(attachment_set['attachTime'], '%Y-%m-%dT%H:%M:%SZ')
            attachment_set['attachTime'] = attach_date
        return attachment_set

    def _to_volumes(self, element):
        element_volumes = findall(element=element, xpath='volumeSet/item', namespace=NAMESPACE)
        return [self._to_volume(volume) for volume in element_volumes]

    def _to_volume(self, element, name=None):
        element_as = findall(element=element, xpath='attachmentSet/item', namespace=NAMESPACE)
        volume = {}
        for key in ['volumeId', 'size', 'createTime', 'status', 'attachmentSet']:
            volume[key] = findtext(element=element, xpath=key, namespace=NAMESPACE)
        if name is None:
            name = volume['volumeId']
        svolume = StorageVolume(id=volume['volumeId'],
                                name=name,
                                size=int(volume['size']),
                                driver=self)
        svolume.extra = {'createTime': datetime.strptime(volume['createTime'],
                                                         '%Y-%m-%dT%H:%M:%S.%fZ'),
                         'status': volume['status'],
                         'attachmentSet': self._get_attachment_set(element_as)}
        return svolume

    def list_volumes(self):
        """
        Eucalyptus specific implementation of list_volumes.
        """
        params = {'Action' : 'DescribeVolumes'}
        element = self.connection.request(self.path, params).object
        volumes = self._to_volumes(element)
        return volumes

    def _build_filter_query(self, query, filter_by):
        """
        Filter syntax found in EC2 API
            Filter.1.Name=owner-alias
            Filter.1.Value.1=amazon
            Filter.1.Value.2=aws-marketplace
            Filter.2.Name=image-type
            Filter.2.Value.1=machine
            ...
        """
        if type(filter_by) is not dict or type(query) is not dict:
            return None
        for (idx,(filter_name,filter_values)) in enumerate(filter_by.items()):
            #Add the new filter by name
            filter_str = 'Filter.%s.Name' % (idx+1)
            query[filter_str] = filter_name
            if type(filter_values) != list:
                filter_values = [filter_values]
            for value_idx, value in enumerate(filter_values):
                #Add each value found for the filter
                filter_str = 'Filter.%s.Value.%s' % (idx+1, value_idx+1)
                query[filter_str] = value
        return query

    def ex_filter_machines(self, filter_by={}):
        """
        Using filter_by:
        #Keys in the dict must match the EC2 API: See "Supported Filters" at http://goo.gl/XNOJv
        #Values in the dict must be str/[str,str,..]
        #EX:
        #Human Readable:
            All windows machines on AWS Marketplace or built by Amazon
        #filter_by=
            {'owner-alias':['amazon','aws-marketplace'], 'image-type':'machine', 'platform':'windows'}
        """
        query = {'Action': 'DescribeImages'}
        self._build_filter_query(query, filter_by)
        images = self._to_images(
            self.connection.request(self.path, query).object
        )
        return images



"""
Atmosphere service machine.

"""
from abc import ABCMeta

from rtwo.models.provider import AWSProvider, EucaProvider, OSProvider

from threepio import logger

class BaseMachine(object):
    __metaclass__ = ABCMeta


class MockMachine(BaseMachine):

    def __init__(self, image_id, provider):
        self.id = image_id
        self.alias = image_id
        self.name = 'Unknown image %s' % image_id
        self._image = None
        self.provider = provider

    def json(self):
        return {'id': self.id,
                'alias': self.alias,
                'name': self.name,
                'provider': self.provider.name}


class Machine(BaseMachine):

    provider = None

    machines = {}

    lc_images = None

    def __init__(self, lc_image):
        self._image = lc_image
        self.id = lc_image.id
        self.alias = lc_image.id
        self.name = lc_image.name

    @classmethod
    def create_machine(cls, provider, lc_image, identifier):
        machine = provider.machineCls(lc_image)
        cls.add_to_cache(provider, machine, identifier)
        return machine

    @classmethod
    def invalidate_provider_cache(cls, provider):
        cls.machines[provider.identifier] = {}

    @classmethod
    def invalidate_machine_cache(cls, provider, machine):
        """
        DO NOT USE THIS METHOD
        Removing a machine from the cache will have no affect on calls to
        driver.list_machines(), to clear the cache you must use
        invalidate_provider_cache
        """
        return cls.invalidate_provider_cache(provider)
        alias = machine.id
        provider_cache = cls.machines.get(provider.identifier, {})
        provider_cache[alias] = None
        cls.machines[provider.identifier] = provider_cache

    @classmethod
    def add_to_cache(cls, provider, machine, identifier):
        alias = machine.id
        machine_dict = cls.machines.get(identifier, {})
        machine_dict[alias] = machine
        cls.machines[identifier] = machine_dict

    @classmethod
    def lookup_cached_machine(cls, alias, identifier):
        provider_machines = cls.machines.get(identifier)
        if not provider_machines:
            #logger.info("Created new machine dict for provider %s" % identifier)
            provider_machines = {}
        machine = provider_machines.get(alias)
        if machine:
            #logger.info("Found machine for provider:%s - %s" %
            #    (identifier, machine))
            return machine
        return None

    @classmethod
    def get_cached_machine(cls, lc_image, identifier):
        alias = lc_image.id
        provider_machines = cls.machines.get(identifier)
        if not provider_machines:
            #logger.info("Created new machine dict for provider %s" % identifier)
            provider_machines = {}
        machine = provider_machines.get(alias)
        if machine:
            #logger.info("Found machine for provider:%s - %s" %
            #    (identifier, machine))
            return machine
        return cls.create_machine(cls.provider, lc_image, identifier)

    @classmethod
    def get_cached_machines(cls, identifier, lc_list_images_method, *args, **kwargs):
        """
        Identifier - Used to identify the specific provider being used:
        ex: "iPlant Eucalyptus", "Openstack 1", "Openstack 2"
        If using only one provider, this variable can be set to any value.
        """
        provider_machines = cls.machines.get(identifier)
        if not provider_machines:
            logger.debug("Cache miss for identifier - %s" % identifier)
        if not provider_machines or not cls.lc_images:
            #Add new provider to the cache
            cls.lc_images = lc_list_images_method(*args, **kwargs)
            logger.debug("Caching %s machines for identifier:%s" %
                         (len(cls.lc_images), identifier))
        return [cls.get_cached_machine(lc_image, identifier) for lc_image in cls.lc_images]

    def reset(self):
        Machine.reset()
        self.machines = {}
        self.lc_images = None

    @classmethod  # order matters... /sigh
    def reset(cls):
        cls.machines = {}
        cls.lc_images = None

    def __unicode__(self):
        return str(self)

    def __str__(self):
        return reduce(
            lambda x, y: x+y,
            map(unicode, [self.__class__, " ", self.json()])
        )

    def __repr__(self):
        return str(self)

    def json(self):
        return {'id': self.id,
                'alias': self.alias,
                'name': self.name,
                'provider': self.provider.name}


class AWSMachine(Machine):

    provider = AWSProvider


class EucaMachine(Machine):

    provider = EucaProvider


class OSMachine(Machine):

    provider = OSProvider



"""
Create a mock driver and attempt to call each mock method that RTwo adds data
to!
"""

import unittest
from mock import Mock, patch

from rtwo.test.secrets import OPENSTACK_PARAMS

from libcloud.utils.py3 import httplib
from libcloud.utils.py3 import method_type
from libcloud.utils.py3 import u
from libcloud.common.types import LibcloudError
from libcloud.compute.types import Provider, KeyPairDoesNotExistError

from libcloud.compute.providers import get_driver
from libcloud.compute.drivers.openstack import (
    OpenStackSecurityGroup, OpenStackSecurityGroupRule,
    OpenStack_1_1_FloatingIpPool, OpenStack_1_1_FloatingIpAddress,
    OpenStackKeyPair
)
from libcloud.test.compute.test_openstack import OpenStack_1_1_MockHttp, \
                                                 OpenStackMockHttp
from libcloud.test.compute.test_openstack import OpenStack_1_1_Tests
from rtwo.drivers.openstack_facade import OpenStack_Esh_Connection,OpenStack_Esh_NodeDriver

######

class OpenStackEshConnectionTest(unittest.TestCase):
    def setUp(self):
        self.timeout = 10
        OpenStack_Esh_Connection.conn_classes = (None, Mock())
        self.connection = OpenStack_Esh_Connection('foo', 'bar',
                                                  timeout=self.timeout,
                                                  ex_force_auth_url='https://127.0.0.1')
        self.connection.driver = Mock()
        self.connection.driver.name = 'OpenStackEshDriver'

    def test_timeout(self):
        self.connection.connect()
        self.assertEqual(self.connection.timeout, self.timeout)
        self.connection.conn_classes[1].assert_called_with(host='127.0.0.1',
                                                           port=443,
                                                           timeout=10)

class OpenStackEshDriverTest(OpenStack_1_1_Tests):
    driver_args = OPENSTACK_PARAMS
    driver_klass = OpenStack_Esh_NodeDriver
    driver_type = OpenStack_Esh_NodeDriver

    def setUp(self):
        super(OpenStackEshDriverTest, self).setUp()





"""
Extension of libcloud's Eucalyptus Node Driver.
"""
import re
import string
from datetime import datetime

from libcloud.utils.xml import fixxpath, findall, findtext, findattr
from libcloud.compute.base import NodeSize, StorageVolume, NodeImage
from libcloud.compute.drivers.ec2 import EucNodeDriver, NAMESPACE

from threepio import logger


class Eucalyptus_Esh_NodeDriver(EucNodeDriver):
    """
    Eucalyptus 2.x node driver for esh.
    """

    def _is_name(self, name):
        for n in ['admin', 'edwins', 'edwin', 'sgregory',
                  'esteve', 'jmatt', 'nirav',
                  'nmatasci', 'sangeeta', 'aedmonds']:
            if n == name:
                return name
        return False

    def _split_name(self, machine_name):
        split = machine_name.split('_')
        length = len(split)
        names = []
        for i in range(length):
            name = split[i]
            if self._is_name(name):
                names.append(name)
            else:
                return (names, ' '.join(split[i:length]))
        return ([], machine_name)

    def _drop_numbers(self, machine_name):
        return '_'.join(filter(lambda x:
                               not str.isdigit(x),
                               machine_name.split('_')))

    def _drop_filetype(self, machine_name):
        new_machine_name = machine_name.replace('.manifest.xml', '')
        return new_machine_name.replace('.img', '')

    def _drop_dir(self, machine_name):
        return machine_name.split('/')[1]

    def parse_machine_name(self, machine_name):
        new_machine_name = self._drop_numbers(
            self._drop_filetype(self._drop_dir(machine_name)))
        names, new_machine_name = self._split_name(new_machine_name)
        #logger.debug('%s->%s' % (machine_name,new_machine_name))
        #logger.debug('names = %s' % names)
        return new_machine_name

    def ex_list_all_instances(self, *args, **kwargs):
        """
        This is merely syntactic sugar, eucalyptus shows all instances when
        the user has the proper access
        """
        return self.list_nodes(*args, **kwargs)

    def list_images(self, location=None, ex_image_ids=None, emi_only=True):
        """
        Ignore any ramdisks/kernels, unless set explicitly
        """
        image_list = super(Eucalyptus_Esh_NodeDriver, self).list_images(
            location, ex_image_ids)
        if emi_only:
            image_list = [img for img in image_list
                          if 'eri' not in img.id and 'eki' not in img.id]
        return image_list

    def _to_image(self, element):
        n = NodeImage(
            id=findtext(element=element, xpath='imageId', namespace=NAMESPACE),
            name=self.parse_machine_name(findtext(element=element,
                                                  xpath='imageLocation',
                                                  namespace=NAMESPACE)),
            driver=self.connection.driver,
            extra={
                'location': findtext(element=element, xpath='imageLocation',
                                                  namespace=NAMESPACE),
                'state': findattr(element=element, xpath="imageState",
                                  namespace=NAMESPACE),
                'ownerid': findattr(element=element, xpath="imageOwnerId",
                                    namespace=NAMESPACE),
                'owneralias': findattr(element=element,
                                       xpath="imageOwnerAlias",
                                       namespace=NAMESPACE),
                'ispublic': findattr(element=element,
                                     xpath="isPublic",
                                     namespace=NAMESPACE),
                'architecture': findattr(element=element,
                                         xpath="architecture",
                                         namespace=NAMESPACE),
                'imagetype': findattr(element=element,
                                      xpath="imageType",
                                      namespace=NAMESPACE),
                'platform': findattr(element=element,
                                     xpath="platform",
                                     namespace=NAMESPACE),
                'kernelid': findattr(element=element,
                                     xpath="kernelId",
                                     namespace=NAMESPACE),
                'ramdiskid': findattr(element=element,
                                      xpath="ramdiskId",
                                      namespace=NAMESPACE),
                'rootdevicetype': findattr(element=element,
                                           xpath="rootDeviceType",
                                           namespace=NAMESPACE),
                'virtualizationtype': findattr(
                    element=element, xpath="virtualizationType",
                    namespace=NAMESPACE),
                'hypervisor': findattr(element=element,
                                       xpath="hypervisor",
                                       namespace=NAMESPACE)
            }
        )
        return n

    def _to_node(self, api_node, groups=None, owner=None):
        def _set_ips(node):
            """
            Set up ips in the return node after _to_node calls it's super.
            """
            pass
            node.public_ips.append(node.extra.get('dns_name', '0.0.0.0'))
            node.private_ips.append(node.extra.get('private_dns', '0.0.0.0'))
            return node
        node = super(Eucalyptus_Esh_NodeDriver, self)._to_node(api_node)
        if owner:
            node.extra['ownerId'] = owner
        node = _set_ips(node)
        return node

    def _get_attachment_set(self, element):
        attachment_set = {}
        for elem in element:
            for key in ['volumeId', 'instanceId', 'device',
                        'status', 'attachTime']:
                value = findtext(element=elem, xpath=key, namespace=NAMESPACE)
                if value:
                    attachment_set[key] = value
        if attachment_set.get('attachTime', None):
            if '.' in attachment_set['attachTime']:
                attach_date = datetime.strptime(attachment_set['attachTime'],
                                                '%Y-%m-%dT%H:%M:%S.%fZ')
            else:
                attach_date = datetime.strptime(attachment_set['attachTime'],
                                                '%Y-%m-%dT%H:%M:%SZ')
            attachment_set['attachTime'] = attach_date
        return attachment_set

    def _to_volume(self, element, name=None):
        element_as = findall(element=element,
                             xpath='attachmentSet/item', namespace=NAMESPACE)
        volume = {}
        for key in ['volumeId', 'size', 'createTime',
                    'status', 'attachmentSet']:
            volume[key] = findtext(element=element,
                                   xpath=key, namespace=NAMESPACE)
        if name is None:
            name = volume['volumeId']
        svolume = StorageVolume(id=volume['volumeId'],
                                name=name,
                                size=int(volume['size']),
                                driver=self)
        try:
            created_time = datetime.strptime(volume['createTime'],
                                             '%Y-%m-%dT%H:%M:%S.%fZ')
        except ValueError:  # Wrong Format, try again..
            created_time = datetime.strptime(volume['createTime'],
                                             '%Y-%m-%dT%H:%M:%SZ')

        svolume.extra = {
            'createTime': created_time,
            'status': volume['status'],
            'attachmentSet': [self._get_attachment_set(element_as)]}
        return svolume

    def _getNextAvailableDevice(self, instance_id):
        """
        Given an instance_id
        return the next available location for attaching a volume
        relative to /dev/
        """
        def _attached_to_instance(vol):
            attached_instance_id = vol.extra.get(
                'attachmentSet', [{}])[0].get('instanceId', '')
            return attached_instance_id == instance_id

        # get all volumes that are attached to this instance
        # add devices to list
        attached_volumes = filter(_attached_to_instance, self.list_volumes())
        used_devices = [vol.extra.get('attachmentSet', [{}])[0].get('device')
                        for vol in attached_volumes]

        logger.debug('List of used devices:%s' % used_devices)

        # start with the letter b, increment until "xvd[char]"
        # is not in the list of used devices
        device_char_index = 1  # that's the index of b. /dev/xvda* is reserved
        letters = string.ascii_letters
        available_character = None
        while available_character is None:
            if '/dev/xvd' + letters[device_char_index] not in used_devices:
                available_character = letters[device_char_index]
            device_char_index = device_char_index + 1

        # Warning: this method will fail if you try to attach > 25 volumes
        # Probably not a typical use case, though
        return "/dev/xvd" + available_character

    def _to_volumes(self, element):
        element_volumes = findall(element=element,
                                  xpath='volumeSet/item', namespace=NAMESPACE)
        volumes = []
        for vol in element_volumes:
            volumes.append(self._to_volume(vol))
        return volumes

    def _to_nodes(self, object, xpath, groups=None, owner=None):
        return [self._to_node(el, groups=groups, owner=owner)
                for el in object.findall(fixxpath(xpath=xpath,
                                                  namespace=NAMESPACE))]

    def list_nodes(self, ex_node_ids=None):
        """
        Specific eucalyptus implementation to include 'owner' in the extra
        """
        params = {'Action': 'DescribeInstances'}
        if ex_node_ids:
            params.update(self._pathlist('InstanceId', ex_node_ids))
        elem = self.connection.request(self.path, params=params).object
        nodes = []
        for rs in findall(element=elem, xpath='reservationSet/item',
                          namespace=NAMESPACE):
            owner = findtext(element=rs, xpath='ownerId', namespace=NAMESPACE)
            groups = [g.findtext('')
                      for g in findall(element=rs,
                                       xpath='groupSet/item/groupId',
                                       namespace=NAMESPACE)]
            nodes += self._to_nodes(rs, 'instancesSet/item', groups, owner)

        nodes_elastic_ips_mappings = self.ex_describe_addresses(nodes)
        for node in nodes:
            ips = nodes_elastic_ips_mappings[node.id]
            node.public_ips.extend(ips)
        return nodes

    def ex_list_all_volumes(self):
        """
        This is merely syntactic sugar, eucalyptus shows all volumes when
        the user has the proper access
        """
        return self.list_volumes()

    def list_volumes(self):
        """
        Eucalyptus specific implementation of list_volumes.
        """
        params = {'Action': 'DescribeVolumes'}
        element = self.connection.request(self.path, params).object
        volumes = self._to_volumes(element)
        return volumes

    def attach_volume(self, node, volume, device=None):
        if device is None:
            device = self._getNextAvailableDevice(node.id)
        result = super(Eucalyptus_Esh_NodeDriver, self).attach_volume(node,
                                                                    volume,
                                                                    device)
        logger.debug('Result of attaching device to %s:%s' % (device, result))
        return result

    def create_volume(self, size, name, location='bespin', *args, **kwargs):
        params = {'Action': 'CreateVolume',
                  'Size': str(size),
                  'AvailabilityZone': location, }
        volume_obj = self.connection.request(self.path, params=params).object
        volume = self._to_volume(volume_obj, name=name)
        logger.debug(volume)
        volume.extra.update({'name': name})
        return True, volume

    def list_sizes(self, location=None):
        """
        Eucalyptus 2 specific implementation of list_sizes.
        """
        params = {'Action': 'DescribeAvailabilityZones',
                  'ZoneName.1': 'verbose'}
        element = self.connection.request(self.path, params).object
        sizes = self._to_sizes(element)
        return sizes

    def _to_sizes(self, element):
        """
        Create a NodeSize object given a libcloud XML element.
        Extra information is available in nodesize.extra dictionary.
        """
        # elem_sizes = element/availabilityZoneInfo/item.
        # elem_sizes[0]  == zone and ip
        # elem_sizes[1]  == column descriptors
        # elem_sizes[2]+ == occupancy and type information.
        elem_sizes = findall(element=element,
                             xpath='availabilityZoneInfo/item',
                             namespace=NAMESPACE)[2:]
        sizes = []
        for s in elem_sizes:
            pieces = re.findall(r'\w+|!/', s[1].text)
            # expected format of pieces is
            # ['0058', '0215', '2', '4096', '10']
            # ['remaining', 'total', 'cpu', 'ram', 'disk']
            id = s[0].text.split()[1]
            size_info = {'id': id,
                         'name': id,
                         'ram': int(pieces[3]),
                         'disk': int(pieces[4]),
                         'bandwidth': 0,
                         'price': 0}
            node_size = NodeSize(driver=self, **size_info)
            node_size.extra = {'cpu': int(pieces[2]),
                               'occupancy': {'remaining': int(pieces[0]),
                                             'total': int(pieces[1])}}
            sizes.append(node_size)
        return sizes



import os

from libcloud.utils.files import read_in_chunks
from libcloud.common.types import LibcloudError
from libcloud.common.base import ConnectionUserAndKey, RawResponse
from libcloud.storage.base import Object, Container, StorageDriver

from libcloud.storage.types import ContainerIsNotEmptyError
from libcloud.storage.types import InvalidContainerNameError
from libcloud.storage.types import ContainerDoesNotExistError
from libcloud.storage.types import ObjectDoesNotExistError
from libcloud.storage.types import ObjectHashMismatchError

from irods.session import iRODSSession
from irods.models import Collection, User, DataObject
from irods.exception import DataObjectDoesNotExist, CollectionDoesNotExist

"""
iRODS data store
Requires 'pycommands' in addition to libcloud
"""


class IRODSConnection(ConnectionUserAndKey):
    """
    iRODS connection that accepts user_id and key
    """

    def __init__(self, username, password, host, port, zone, client_user=None,
                 client_zone=None, **kwargs):
        self.user_id = username
        self.key = password
        self.host = host
        self.port = port
        self.zone = zone
        if client_user:
            self.client_user = client_user
            self.client_zone = client_zone if client_zone else zone


    def connect(self, host=None, port=None, base_url=None, client_user=None,
            client_zone=None):
        if not host:
            host = self.host
        if not port:
            port = self.port
        session = iRODSSession(host=host, port=port,
                                  user=self.user_id, password=self.key,
                                  zone=self.zone, client_user=client_user,
                                  client_zone=client_zone)
        self.session = session
        return session

    def listdir(self, collection):
        """
        Expects an iRODSCollection
        """
        files = collection.data_objects
        dirs = collection.subcollections
        return (dirs, files)

    def walk(self, root_fs, topdown=True, onerror=None):
        """Directory tree generator.

        For each directory in the directory tree rooted at top (including top
        itself, but excluding '.' and '..'), yields a 3-tuple

            dirpath, dirnames, filenames
        """
        try:
            top = self.session.collections.get(root_fs)
            dirs, nondirs = self.listdir(top)
        except Exception, err:
            if onerror is not None:
                onerror(err)
            raise
            #return

        if topdown:
            yield top, dirs, nondirs
        for subcollection in dirs:
            new_path = subcollection.path
            for x in self.walk(new_path, topdown, onerror):
                yield x
        if not topdown:
            yield top, dirs, nondirs
    
class IRODSDriver(StorageDriver):
    name = 'iRODS Data Store'
    website = 'http://www.irods.org/'
    hash_type = 'md5'
    connectionCls = IRODSConnection

    def __init__(self, key, secret=None, host=None, port=None, zone=None,
                 client_user=None, client_zone=None, base_path=None, **kwargs):
        self.key = key
        self.secret = secret
        self.zone = zone

        args = [self.key]

        if self.secret is not None:
            args.append(self.secret)
        if host is not None:
            args.append(host)
        if port is not None:
            args.append(port)
        if self.zone is not None:
            args.append(self.zone)
        if not base_path:
            base_path = '/%s/home/%s' % (self.zone, self.key)
        self.base_path = base_path

        self.connection = self.connectionCls(*args,
            **self._ex_connection_class_kwargs())

        self.connection.driver = self
        self.connection.connect(client_user=client_user,
                                client_zone=client_zone)

    def iterate_containers(self):
        """
        Return a generator of containers.

        @return: A generator of Container instances.
        @rtype: C{generator} of L{Container}
        """
        base_container = self.connection.session.collections.get(self.base_path)
        if not base_container:
            raise LibcloudError(
                'Failed to retrieve containers from base container:%s' % self.base_path,
                driver=self)
        for collection in base_container.subcollections:
            yield self._to_container(collection)

    def list_containers(self):
        """
        Return a list of containers.

        @return: A list of Container instances.
        @rtype: C{list} of L{Container}
        """
        base_container = self.connection.session.collections.get(self.base_path)
        if not base_container:
            raise LibcloudError(
                'Failed to retrieve containers from base container:%s' % self.base_path,
                driver=self)
        containers = self._to_containers(base_container.subcollections)
        return containers

    def list_container_objects(self, container, ex_prefix=None):
        """
        Return a list of objects for the given container.

        :param container: Container instance.
        :type container: :class:`Container`

        :param ex_prefix: Only return objects starting with ex_prefix
        :type ex_prefix: ``str``

        :return: A list of Object instances.
        :rtype: ``list`` of :class:`Object`
        """
        return self._get_objects(container)

    def iterate_container_objects(self, container, ex_prefix=None):
        """
        Return a generator of objects for the given container.

        :param container: Container instance
        :type container: :class:`Container`

        :param ex_prefix: Only return objects starting with ex_prefix
        :type ex_prefix: ``str``

        :return: A generator of Object instances.
        :rtype: ``generator`` of :class:`Object`
        """
        #TODO: Filter on ex_prefix
        return iter(self._get_objects(container))

    def get_container(self, container_path):
        try:
            collection = self.connection.session.collections.get(container_path)
            return self._to_container(collection)
        except CollectionDoesNotExist:
            raise ContainerDoesNotExistError(value=None,
                                             driver=self,
                                             container_name=container_path)

    def get_object(self, container_path, object_name):
        try:
            container = self.get_container(container_path)
        except CollectionDoesNotExist:
            raise ContainerDoesNotExistError(value=None, driver=self,
                                             container_name=container_path)
        try:
            obj_path = self._get_object_path(container, object_name)
            data_obj = self.connection.session.data_objects.get(obj_path)
            return self._to_obj(data_obj, container)
        except DataObjectDoesNotExist:
            raise ObjectDoesNotExistError(value=None, driver=self,
                                          object_name=object_name)


    def create_object(self, container, object_name):
        """
        Create a new object 'object_name' inside container
        """
        object_path = self._get_object_path(obj.container, obj.name)
        try:
            new_data_object = self.connection.session.data_objects.create(object_path)
            return self._to_obj(new_data_object)
        except: #CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME
            raise ObjectError(value='Object with this name already exists.'
                                    'Name must be unique among all objects '
                                    'in this container',
                              driver=self, object_name=object_name)

    def create_container(self, container_name):
        """
        Create a new container with path 'container_name'
        """
        #Containers do not end in slash
        if container_name.endswith('/'):
            container_name = container_name[:-1]

        self._check_container_name(container_name)
        try:
            new_collection = self.connection.session.collections.create(container_name)
            return self._to_container(new_collection)
        except: #CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME
            raise InvalidContainerNameError(
                value='Container with this name already exists. The name must '
                      'be unique among all the containers in the system',
                container_name=container_name, driver=self)

    def delete_container(self, container):
        # Note: All the objects in the container must be deleted first
        
        for obj in self._get_objects(container):
            raise ContainerIsNotEmptyError(value='Container is not empty',
                                container_name=container.name, driver=self)
        
        try:
            self.connection.session.collections.delete(container.name)
            return True
        except: #CollectionDoesNotExist
            raise ContainerDoesNotExistError(value=None,
                                             driver=self,
                                             container_name=container.name)


    def download_object(self, obj, destination_path, overwrite_existing=False,
                        delete_on_failure=True):
        """
        Download an object to the specified destination path.

        @param obj: Object instance.
        @type obj: L{Object}

        @param destination_path: Full path to a file or a directory where the
                                incoming file will be saved.
        @type destination_path: C{str}

        @param overwrite_existing: True to overwrite an existing file,
            defaults to False.
        @type overwrite_existing: C{bool}

        @param delete_on_failure: True to delete a partially downloaded file if
        the download was not successful (hash mismatch / file size).
        @type delete_on_failure: C{bool}

        @return: True if an object has been successfully downloaded, False
        otherwise.
        @rtype: C{bool}
        """
        #Check that download location exists
        base_name = os.path.basename(destination_path)

        if not base_name and not os.path.exists(destination_path):
            raise LibcloudError(
                value='Path %s does not exist' % (destination_path),
                driver=self)
        #Determine the file_path where file will be saved
        if not base_name:
            file_path = os.path.join(destination_path, obj.name)
        else:
            file_path = destination_path
        #Check that the object/container exists
        try:
            from_dir = obj.container.name
            collection = self.connection.session.collections.get(from_dir)
        except: #CollectionDoesNotExist
            raise LibcloudError(
                value='Container %s does not exist,' % (base_container) +
                'cannot download object to %s' % (destination_path),
                driver=self)

        if os.path.exists(file_path) and not overwrite_existing:
            raise LibcloudError(
                value='File %s already exists, but ' % (file_path) +
                'overwrite_existing=False',
                driver=self)

        with obj.extra['irods'].open('r+') as read_file:
            with open(file_path, 'w') as write_file:
                for line in read_file:
                    write_file.write(line)

    def download_object_as_stream(self, obj, chunk_size=None):
        """
        Return a generator which yields object data.

        @param obj: Object instance
        @type obj: L{Object}

        @param chunk_size: Optional chunk size (in bytes).
        @type chunk_size: C{int}

        @rtype: C{object}
        """
        #Fails with: 'iRODSDataObjectFile' object has no attr 'next'
        #with obj.extra['irods'].open('r+') as read_file:
        #    for data in read_in_chunks(read_file, chunk_size=chunk_size):
        #        yield data
        raise NotImplementedError()

    def upload_object(self, file_path, container, object_name, extra=None,
                      verify_hash=True):
        """
        Upload an object currently located on a disk.

        @param file_path: Path to the object on disk.
        @type file_path: C{str}

        @param container: Destination container.
        @type container: L{Container}

        @param object_name: Object name.
        @type object_name: C{str}

        @param verify_hash: Verify hash
        @type verify_hash: C{bool}

        @param extra: (optional) Extra attributes (driver specific).
        @type extra: C{dict}

        @rtype: C{object}
        """
        object_path = self._get_object_path(container, object_name)
        try:
            data_object = self.connection.session.data_objects.get(object_path)
        except DataObjectDoesNotExist:
            #Create the object
            data_object = self.connection.session.data_objects.create(object_path)

        with data_object.open('w') as f_to_write:
            with open(file_path,'r') as f_to_read:
                for line in f_to_read:
                    f_to_write.write(line)
        #Grab the newly updated object (After the write)
        data_object = self.connection.session.data_objects.get(object_path)
        return self._to_obj(data_object, container)

    def delete_object(self, obj):
        object_path = self._get_object_path(obj.container, obj.name)
        try:
            self.connection.session.data_objects.unlink(object_path)
            return True
        except DataObjectDoesNotExist:
            raise ObjectDoesNotExistError(value=None, driver=self,
                                          object_name=obj.name)


    def _get_objects(self, container):
        """
        Iterate through all data_objects on iRODS container
        """
        container_url = self._get_container_path(container)
        for root, collections, data_objects in self.connection.walk(container_url):
            for data_obj in data_objects:
                yield self._to_obj(data_obj, root)

    def _get_container_path(self, container):
        """
        Return a container path

        @param container: Container instance
        @type  container: L{Container}

        @return: A path for this container.
        @rtype: C{str}
        """
        #NOTE: Container.name contains FULL PATH
        return '%s' % (container.name)

    def _get_object_path(self, container, object_name):
        """
        Return path of object on iRODS File System

        @param container: Container instance
        @type  container: L{Container}

        @param object_name: Object name
        @type  object_name: L{str}

        @return: A  path for this object.
        @rtype: C{str}
        """
        container_url = self._get_container_path(container)
        object_name_cleaned = self._clean_object_name(object_name)
        object_path = '%s/%s' % (container_url, object_name_cleaned)
        return object_path

    def _clean_object_name(self, name):
        """
        These rules define 'allowable' names.
        """
        name = name.replace('\\','').replace('/','').replace(' ','_')
        return name

    def _check_container_name(self, container_name):
        """
        Check if the container name is valid

        @param container_name: Container name
        @type container_name: C{str}
        """

        if ' ' in container_name:
            raise InvalidContainerNameError(value=None, driver=self,
                                            container_name=container_name)

    def _to_containers(self, subcollections):
        for collection in subcollections:
            yield self._to_container(collection)

    def _to_objs(self, obj, container):
        return [self._to_obj(element, container) for element in obj]

    def _to_container(self, irods_collection):
        extra = {
            'name': irods_collection.name,
            'path': irods_collection.path,
            'id': irods_collection.id,
            'meta': irods_collection.metadata._meta
        }

        container = Container(name=irods_collection.path, 
                              extra=extra, driver=self)
        return container

    def _to_obj(self, irods_data_obj, container):
        extra = {'path': irods_data_obj.path,
                 'create_time': irods_data_obj.create_time,
                 'irods': irods_data_obj,
                 'last_modified': irods_data_obj.modify_time}
        obj = Object(name=irods_data_obj.name,
                     size=irods_data_obj.size,
                     hash=irods_data_obj.checksum,
                     extra=extra,
                     meta_data={},
                     container=container,
                     driver=self
                     )

        return obj



"""
Instance Link Testing.
"""
import multiprocessing

import requests

from threepio import logger


class LinkTestProcess(multiprocessing.Process):
    def __init__(self, task_queue, result_queue):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue

    def run(self):
        #proc_name = self.name
        while True:
            #Do work
            next_task = self.task_queue.get()
            if next_task is None:
                break
            result = next_task()
            self.result_queue.put(result)
        return


class LinkTestTask():
    def __init__(self, alias, uri):
        self.alias = alias
        self.uri = uri

    def __call__(self):
        link_results = test_instance_links(self.alias, self.uri)
        return link_results

    def __str__(self):
        return "%s" % (self.instance)


def test_instance_links(alias, uri):
    #logger.debug(uri)
    shell_address = 'http://%s:4200' % uri
    shell_success = test_link(shell_address)
    vnc_address = 'http://%s:5904' % uri
    vnc_success = test_link(vnc_address)
    return {alias: {'vnc': vnc_success, 'shell': shell_success}}


def test_link(address):
    if not address:
        return False
    try:
        response = requests.head(address, timeout=9.0)
        if response.status_code in [200, 302]:
            return True
        return False
    except requests.ConnectionError, error:
        err_code, err_reason = error.args[0].reason
        logger.warn("Link test failed: URL:%s Error:%s - %s" % (address, err_code, err_reason))
        return False
    except Exception as e:
        logger.exception(e)
        return False

def active_instances(instances):
    return active_instances_naive(instances)

def active_instances_naive(instances):
    test_results = {}
    for instance in instances:
        if instance.ip is not None and instance.extra['status'] == 'active':
            link_results = test_instance_links(instance.alias, instance.ip)
        else:
            logger.info("Not testing %s:%s-%s" % (instance,
                                               instance.ip,
                                               instance.extra['status']))
        test_results.update(link_results)
    return test_results

def active_instances_threaded(instances):
    """
    Creates multiple processes to test instance links
    """
    test_results = {}

    # Determine #processes and #jobs
    num_processes = multiprocessing.cpu_count() * 2
    num_jobs = len(instances)
    # logger.debug("Created %d processes to run %s jobs" %
    #             (num_processes, num_jobs))

    # Create input and output queue
    tasks = multiprocessing.Queue()
    results = multiprocessing.Queue()

    processes = [LinkTestProcess(tasks, results)
                 for i in xrange(num_processes)]
    for p in processes:
        p.start()
        # logger.info("Started %d processes" % (num_processes,))

    for i in instances:
        # Task to run on idle processes
        tasks.put(LinkTestTask(i.alias, i.ip))
        # logger.info("Added %d tasks" % (num_jobs,))

    for _ in xrange(num_processes):
        # Sentinal value to kill the proc
        tasks.put(None)
        # logger.info("Added %d poison pills" % (num_processes,))

    while num_jobs:
        # logger.info("in num_jobs")
        # logger.info(results)
        try:
            result = results.get()
            if result:
                test_results.update(result)
        except Exception:
            logger.exception("Problem with multiprocessing queue.")
        num_jobs -= 1

    #logger.info("Threads complete. Returning response")
    return test_results



"""
Rtwo mixins for drivers
Mixin classes implement additional functionality for Drivers.
"""
from threepio import logger

class MetaMixin():
    def meta(self, *args, **kwargs):
        return self.provider.metaCls.create_meta(self, *args, **kwargs)


class APIFilterMixin():
    """
    APIFilterMixin provides filtering for libcloud and esh drivers.
    """
    def get_volume(self, alias):
        try:
            volume_list = self.list_volumes()
            volume = filter(lambda volume:
                            alias == volume.alias, volume_list)[0]
            return volume
        except IndexError:
            return None

    def get_size(self, alias):
        try:
            size_list = self.list_sizes()
            size = filter(lambda size:
                          alias == size.id, size_list)[0]
            return size
        except IndexError:
            return None

    def get_instance(self, alias):
        try:
            instance_list = self.list_instances()
            instance = filter(lambda instance:
                              alias == instance.alias, instance_list)[0]
            return instance
        except IndexError:
            return None

    def get_machine(self, alias):
        try:
            machine_list = self.list_machines()
            machine = filter(lambda machine:
                             alias == machine.alias, machine_list)[0]
            return machine
        except IndexError:
            return None

    def filter_volumes(self, volumes, black_list=[]):
        """
        Filtered volumes:
            Keep the volume if it does NOT match any word in the black_list
        """
        filtered = [volume for volume in volumes
                    if not any(word in volume.name for word in black_list)]
        return filtered

    def filter_sizes(self, sizes, black_list=[]):
        """
        Filtered sizes:
            Keep the size if it does NOT match any word in the black_list
        """
        filtered = [size for size in sizes
                    if not any(word in size.name for word in black_list)]
        return filtered

    def filter_instances(self, instances, black_list=[]):
        """
        Filtered instances:
            Keep the instance if it does NOT match any word in the black_list
        """
        filtered = [instance for instance in instances
                    if not any(word in instance.name for word in black_list)]
        return filtered

    def filter_machines(self, machines, black_list=[]):
        """
        Filtered machines:
            Keep the machine if it does NOT match any word in the black_list
        """
        filtered = [machine for machine in machines
                    if not any(word in machine.name for word in black_list)]
        return filtered


class InstanceActionMixin():
    def reboot_instance(self, *args, **kwargs):
        raise NotImplementedError

    def resume_instance(self, *args, **kwargs):
        raise NotImplementedError

    def suspend_instance(self, *args, **kwargs):
        raise NotImplementedError

    def resize_instance(self, *args, **kwargs):
        raise NotImplementedError

    def confirm_resize_instance(self, *args, **kwargs):
        raise NotImplementedError

    def revert_resize_instance(self, *args, **kwargs):
        raise NotImplementedError



""" rtwo.settings

    Try to find usable settings. First try to use atmosphere's secrets,
    if installed. Otherwise try to use test_settings.
"""
import threepio
if not threepio.logger:
    threepio.initialize("rtwo")
from threepio import logger

try:
    # Attempt initialization from atmosphere (Using secrets)
    from atmosphere.settings import secrets as settings
    dir(settings)  # Force it to fail.
except:
    settings = None
    try:
        from rtwo import local_settings as settings
        dir(settings)  # Force it to fail.
    except:
        try:
            from rtwo import test_settings as settings
            dir(settings)  # Force it to fail.
        except:
            settings = None


if not settings:
    class Settings(object):
        pass
    settings = Settings()
    settings.EUCA_ADMIN_KEY = ""
    settings.EUCA_ADMIN_SECRET = ""
    settings.SERVER_URL = ""
    settings.INSTANCE_SERVICE_URL = ""
    settings.ATMOSPHERE_VNC_LICENSE = ""


def set_settings(settings):
    global EUCA_ADMIN_KEY, EUCA_ADMIN_SECRET, SERVER_URL,\
        INSTANCE_SERVICE_URL, ATMOSPHERE_VNC_LICENSE
    EUCA_ADMIN_KEY = settings.EUCA_ADMIN_KEY
    EUCA_ADMIN_SECRET = settings.EUCA_ADMIN_SECRET
    SERVER_URL = settings.SERVER_URL
    INSTANCE_SERVICE_URL = settings.INSTANCE_SERVICE_URL
    ATMOSPHERE_VNC_LICENSE = settings.ATMOSPHERE_VNC_LICENSE

set_settings(settings)



"""
Atmosphere service instance.

"""
from threepio import logger

from rtwo.models.provider import AWSProvider, EucaProvider, OSProvider
from rtwo.models.volume import OSVolume, Volume, MockVolume
from rtwo.models.machine import OSMachine, Machine, MockMachine
from rtwo.models.size import Size, MockSize


class Instance(object):

    owner = None
    provider = None
    source = None
    machine = NotImplementedError(
            "This field is deprecated. Use 'source' instead")
    size = None

    def _get_source_for_instance(self, node):
        """
        Retrieve correct source based on instance details
        NOTE: Occasionally more data may be required/things may slow down here.
        """
        source = self._get_source_volume(node)
        if source:
            return source
        source = self._get_source_snapshot(node)
        if source:
            return source
        source = self._get_source_image(node)
        return source

    def _get_source_snapshot(self, node):
        return None
    def _get_source_image(self, node):
        return None
    def _get_source_volume(self, node):
        return None


    def __init__(self, node, provider):

        self.owner = None # Should be defined per-provider
        self._node = node
        self.id = node.id
        self.alias = node.id
        self.name = node.name
        self.extra = node.extra
        self.provider = provider
        self.ip = self.get_public_ip()
        self.source = self._get_source_for_instance(node)

    @classmethod
    def get_instances(cls, nodes, provider):
        return [cls.provider.instanceCls(node, provider) for node in nodes]

    def get_public_ip(self):
        raise NotImplementedError()

    def get_status(self):
        raise NotImplementedError()

    def load(self):
        raise NotImplementedError()

    def save(self):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    def reset(self):
        self._node = None

    def __unicode__(self):
        return str(self)

    def __str__(self):
        return reduce(
            lambda x, y: x+y,
            map(unicode, [self.__class__, " ", self.json()]))

    def __repr__(self):
        return str(self)

    #Marked for deletion - SG
    def json(self):
        size_str = None
        source_str = None
        if not self.size:
            size_str = "None"
        elif type(self.size) == str:
            size_str = self.size
        else:
            size_str = self.size.json()
        if not self.source:
            source_str = "None"
        else:
            source_str = self.source.json()

        return {'id': self.id,
                'alias': self.alias,
                'name': self.name,
                'ip': self.ip,
                'provider': self.provider.name,
                'size': size_str,
                'source': source_str
            }


class AWSInstance(Instance):

    provider = AWSProvider

    def __init__(self, node, provider):
        Instance.__init__(self, node, provider)
        self.owner = node.extra.get('ownerId')
        self.size = node.extra.get('instance_type')
        if not self.size:
            self.size = node.extra['instancetype']
        if Size.sizes.get((provider.identifier, self.size)):
            self.size = Size.sizes[(provider.identifier, self.size)]

    def get_public_ip(self):
        if self.extra \
           and self.extra.get('dns_name'):
            return self.extra['dns_name']


class EucaInstance(AWSInstance):

    provider = EucaProvider

    def get_public_ip(self):
        if self.extra:
            return self.extra.get('dns_name')

    def get_status(self):
        """
        """
        status = "Unknown"
        if self.extra \
           and self.extra.get('status'):
            status = self.extra['status']
        return status


class OSInstance(Instance):

    provider = OSProvider

    def __init__(self, node, provider):
        Instance.__init__(self, node, provider)

        #Unfortunately we can't get the tenant_name..
        self.owner = node.extra.get('tenantId')
        #New in 0.2.11 - use MockSize and expect user to lookup size.id if they want more than a MockSize!
        if not self.size:
            self.size = self._get_flavor_for_instance(node)

    def _check_for_volumes(self, node):
        booted_volume = node.extra['object'].get('os-extended-volumes:volumes_attached')
        if type(booted_volume) == list and booted_volume:
            return booted_volume[0].get('id')
        elif booted_volume:
            #As-of-now this is an impossible situation (Always a list)
            # but in case things change, this is an easy fallback..
            return booted_volume.get('id')
        return None

    def _get_source_volume(self, node):
        """
        Returns None or OSVolume
        """
        #SAFETY check until we can lessen the impact on calls
        if 'bootable_volume' not in node.extra['metadata']:
            return None
        volume_id = self._check_for_volumes(node)
        if not volume_id:
            return None
        volume = self._test_node_is_booted_volume(node, volume_id)
        if not volume:
            return None
        source = OSVolume(volume)
        return source

    def _test_node_is_booted_volume(self, node, volume_id):
        """
        Given a node and a volume_id, return 'volume' if the node
        is 'running' the volume, otherwise return None 
        """
        volume = node.driver.ex_get_volume(volume_id)
        if not volume:
            logger.warn("[BADDATA] Volume %s listed in 'attached_volumes' but did not"
                        " return a volume" % volume_id)
            return None
        attachment_data = volume.extra['attachments'][0]
        if not attachment_data:
            logger.warn("[BADDATA] Volume %s listed in 'attached_volumes' but did not"
                        " return attachment data." % volume_id)
            return None
        device = attachment_data.get("device")
        if device and 'vda' in device:
            return volume
        #Normal behavior, this is NOT a booted volume.
        logger.debug("Volume %s listed in 'attached_volumes' but is NOT "
                     "currently running as an instance." % volume_id)
        return None

    def _get_source_image(self, node):
        """
        NOTE: Always returns a correct source
        That source may be a 'MockMachine' or an 'OSMachine'
        """
        image_id = node.extra.get('imageId')
        if not image_id:
            image_id = node.extra.get('image_id')
        if not image_id:
            return None
        machine = self.provider.machineCls.lookup_cached_machine(image_id,
                self.provider.identifier)
        if not machine:
            machine = MockMachine(node.extra['imageId'], self.provider)
        return machine

    def _get_flavor_for_instance(self, node):
        #Step 1, pure-cache lookup
        size = self.provider.sizeCls.lookup_size(node.extra['flavorId'],
                self.provider)
        if not size:
            size = MockSize(node.extra['flavorId'], self.provider)
        return size

    def get_status(self):
        """
        TODO: If openstack: Use extra['task'] and extra['power']
        to determine the appropriate status.
        """
        status = "Unknown"
        if self.extra \
           and self.extra.get('status'):
            status = self.extra['status']
            task = self.extra.get('task')
            if task:
                status += ' - %s' % self.extra['task']
            extra_status = self.extra.get('metadata', {}).get('tmp_status')
            if extra_status and not task and status == 'active':
                status += ' - %s' % extra_status

        return status

    def get_public_ip(self):
        if hasattr(self, "ip"):
            return self.ip
        if self._node and self._node.public_ips:
            return self._node.public_ips[0]



"""
OpenStack CloudAdmin Libarary
    Use this library to:
    * manage users within Keystone - openstack auth
"""
import os

from keystoneclient.exceptions import NotFound, ClientException
from keystoneclient.exceptions import Conflict as KeystoneConflict
from novaclient.exceptions import OverLimit
from novaclient.exceptions import NotFound as NovaNotFound

from threepio import logger

from rtwo.drivers.common import _connect_to_keystone,\
    _connect_to_swift, _connect_to_nova, find


class UserManager():
    keystone = None
    nova = None
    user = None
    password = None
    project = None

    @classmethod
    def lc_driver_init(self, lc_driver, *args, **kwargs):
        lc_driver_args = {
            'username': lc_driver.key,
            'password': lc_driver.secret,
            'tenant_name': lc_driver._ex_tenant_name,
            'auth_url': lc_driver._ex_force_auth_url.replace('/tokens',''),
            'region_name': lc_driver._ex_force_service_region
        }
        lc_driver_args.update(kwargs)
        manager = UserManager(*args, **lc_driver_args)
        return manager

    def __init__(self, *args, **kwargs):
        self.keystone, self.nova, self.swift = self.new_connection(*args, **kwargs)
        auth_version = kwargs.get('version','v2.0')
        if '2.0' in auth_version:
            self.version = 2
        elif '3' in auth_version:
            self.version = 3
        else:
            raise Exception("Could not determine a version of openstack_user based on %s" % auth_version)


    def new_connection(self, *args, **kwargs):
        keystone = _connect_to_keystone(*args, **kwargs)
        nova_args = kwargs.copy()
        #HACK - Nova is certified-broken-on-v3. 
        nova_args['version'] = 'v2.0'
        nova_args['auth_url'] = nova_args['auth_url'].replace('v3','v2.0')
        nova = _connect_to_nova(*args, **nova_args)
        swift_args = self._get_swift_args(*args, **kwargs)
        swift = _connect_to_swift(*args, **swift_args)
        return keystone, nova, swift

    def _get_swift_args(self, *args, **kwargs):
        swift_args = {}
        swift_args['authurl'] = kwargs.get('auth_url')
        swift_args['user'] = kwargs.get('username')
        swift_args['key'] = kwargs.get('password')
        swift_args['tenant_name'] = kwargs.get('tenant_name')
        if "v2" in kwargs.get("auth_url"):
            swift_args['auth_version'] = 2
        swift_args['os_options'] = {"region_name":kwargs.get("region_name")}
        return swift_args


    def build_nova(self, username, password, project_name, *args, **kwargs):
        """
        Ocassionally you will need the 'user nova' instead of admin nova.
        This function will build a 'proper' nova given credentials.
        """
        converted_kwargs = {
            'username':username,
            'password':password,
            'tenant_name':project_name,
            'auth_url':self.nova.client.auth_url,
            'region_name':self.nova.client.region_name}
        #TODO: Update w/ kwargs..
        nova = _connect_to_nova(*args, **converted_kwargs)
        nova.client.region_name = self.nova.client.region_name
        return nova

    def include_admin(self, projectname, admin_rolename='admin'):
        """
        This should be called each time an account is created:
          This gives the admin tenant access to view the users resources
          This is REQUIRED for:
          * Monitoring of instances
          * Ops
        """
        try:
            # keystone admin always gets access, always has admin priv.
            admin_role_created = self.add_project_membership(
                projectname, self.keystone.username, admin_rolename)
            return admin_role_created
        except ClientException:
            logger.warn('Could not assign admin role to username %s' %
                        self.keystone.username)
            return None

    def build_security_group(self, username, password, project_name,
             security_group_name, protocol_list, rebuild=False, *args, **kwargs):
        """
        Given a working set of credentials and a list of protocols/rules
          Protocol/Rule: ("TCP/UDP/ICMP", from_port, to_port[, CIDR_to_allow])

        Retrieve existing security group and ensure each rule is created if it
        doesnt already exist.

        rebuild: If True, delete all rules before adding rules in protocol_list
        """
        sec_group = self.find_security_group(
                username, password, project_name, security_group_name)

        if not sec_group:
            raise Exception("No security group named %s found for %s"
                        % (security_group_name, username))

        if rebuild:
            rule_ids = [rule['id'] for rule in sec_group.rules]
            self.delete_security_group_rules(
                username, password, project_name, security_group_name, rule_ids)


        # Add the rule, grab updated security group
        self.add_security_group_rules(
            username, password, project_name, security_group_name, protocol_list)

        #Show the new security group (with added rules)
        sec_group = self.find_security_group(
                username, password, project_name, security_group_name)
        return sec_group

    def delete_security_group_rules(self, username, password, project_name,
            security_group_name, rules):
        """
        rules - a list of rules in the form:
                [rule_id1, rule_id2, rule_id3, ...]
        """
        sec_group = self.find_security_group(username, password,
                                             project_name, security_group_name)
        nova = self.build_nova(username, password, project_name)
        for rule_id in rules:
            try:
                nova.security_group_rules.delete(rule_id)
            except NovaNotFound:
                pass

    def add_security_group_rules(self, username, password, project_name,
            security_group_name, rule_list):
        """
        rules - a list of rules in the form:
                [(protocol, from_port, to_port, [CIDR]),
                 ...]
        """
        sec_group = self.find_security_group(username, password,
                                             project_name, security_group_name)
        if not sec_group:
            raise Exception("Could not find security group %s for project %s"
                            % (security_group_name, project_name))
        nova = self.build_nova(username, password, project_name)
        for protocol in rule_list:
            self.add_rule_to_group(nova, protocol, sec_group)

    def add_rule_to_group(self, nova, protocol, security_group):
        """
        Add a security group rule if it doesn't already exist.
        """
        if len(protocol) == 3:
            (ip_protocol, from_port, to_port) = protocol
            cidr = None
        elif len(protocol) == 4:
            (ip_protocol, from_port, to_port, cidr) = protocol
        else:
            raise Exception("Rule tuple did not match expected output:"
                            " (protocol, from_port, to_port, [CIDR])")
        if not self.find_rule(security_group, ip_protocol,
                          from_port, to_port):
            try:
                nova.security_group_rules.create(security_group.id,
                                                 ip_protocol=ip_protocol,
                                                 from_port=from_port,
                                                 to_port=to_port,
                                                 cidr=cidr)
            except OverLimit, ole:
                if 'Security group rule already exists' in ole.message:
                    return True
                logger.exception(ole.__dict__)
                raise
        return True


    def find_security_group(self, username, password, project_name,
                            security_group_name):
        nova = self.build_nova(username, password, project_name)
        try:
            sec_group = nova.security_groups.find(description=security_group_name)
            return sec_group
        except NovaNotFound:
            sec_group = nova.security_groups.list()
        if sec_group:
            return sec_group[0]
        else:
            raise Exception("Could not find any security groups for user %s"
                            % username)


    def list_security_groups(self, username, password, project_name):
        nova = self.build_nova(username, password, project_name)
        sec_groups = nova.security_groups.list()
        return sec_groups

    def find_rule(self, security_group, ip_protocol,
                  from_port, to_port, cidr=None):
        ip_protocol = ip_protocol.lower()
        for rule in security_group.rules:
            if rule['ip_protocol'] and rule['ip_protocol'].lower() == ip_protocol\
                    and (not cidr or rule['ip_range']['cidr'] == cidr):
                #Potential match
                if rule['from_port'] <= from_port\
                        and rule['to_port'] >= to_port:
                    return True
                if rule['from_port'] == None\
                        and rule['to_port'] >= to_port:
                            # The ICMP rule on grizzly-openstack is:
                            # {'from_port': -1, ...} but on havana it is:
                            # {'from_port': None, ...} so check for both
                    return True
        return False

    def add_usergroup(self, username, password, createUser=True, adminRole=False):
        """
        Create a group for this user only
        then create the user
        TODO: drop createUser -- ignored!
        """
        #Create user
        try:
            user = self.create_user(username, password, username)
        except ClientException as user_exists:
            logger.debug('Received Error %s on add, User exists.' %
                         user_exists)
            user = self.get_user(username)

        logger.debug("Assign project:%s Member:%s Role:%s" %
                    (username, username, adminRole))

        #Create project for user/group
        project = self.get_project(username)
        if not project:
            project = self.create_project(username)

        # Check the user has been given an appropriate role
        admin_role_name = "admin"
        if not adminRole:
            role_name = "_member_"
        else:
            role_name = admin_role_name
        try:
            role = self.add_project_membership(username, username, role_name)
        except ClientException:
            logger.warn('Could not assign role to username %s' % username)
        self.include_admin(username, admin_role_name)
        return (project, user, role)

    def get_usergroup(self, username):
        return self.get_project(username)

    ##ADD##
    def create_role(self, rolename):
        """
        Create a new role
        """
        return self.keystone.roles.create(name=rolename)

    def create_project(self, groupname, domain='default'):
        """
        Create a new project
        """
        try:
            if self.keystone_version() == 2:
                return self.keystone_projects().create(groupname)
            else:
                return self.keystone_projects().create(groupname, domain)
        except Exception, e:
            logger.exception(e)
            raise

    def add_project_membership(self, groupname, username, rolename, domain_name=None):
        """
        Adds user(name) to group(name) with role(name)

        Invalid groupname, username, rolename :
            raise keystoneclient.exceptions.NotFound
        """
        # Check for previous entry
        existing_grant = self.check_membership(groupname, username, rolename, domain_name)
        if existing_grant:
            return existing_grant
        # Create a new entry
        try:
            user_kwargs = project_kwargs = {}
            if domain_name and self.keystone_version() != 2:
                user_kwargs.update({'domain_id':domain_name})
                project_kwargs.update({'domain_id':domain_name})
            user = self.get_user(username, **user_kwargs)
            project = self.get_project(groupname, **project_kwargs)
            new_role = self.get_role(rolename)
            if self.keystone_version() == 3:
                role_grant = self.keystone.roles.grant(
                    user=user, project=project, role=new_role)
            else:
                role_grant = project.add_user(user, new_role)
            return role_grant
        except KeystoneConflict, keystone_err:
            if 'already has role' in keystone_err.message:
                return None
            raise
        except Exception, e:
            logger.exception(e)
            raise

    def check_membership(self, projectname, username, rolename, domain_name=None):
        user_kwargs = project_kwargs = {}
        if domain_name:
            user_kwargs.update({'domain_id':domain_name})
            project_kwargs.update({'domain_id':domain_name})
        user = self.get_user(username, **user_kwargs)
        project = self.get_project(projectname, **project_kwargs)
        if not user or not project:
            return None
        new_role = self.get_role(rolename)

        # Check for previous entries
        if self.keystone_version() == 2:
            existing_roles = user.list_roles(project)
        elif self.keystone_version() == 3:
            existing_roles = self.keystone.roles.list(user=user, project=project)

        for role in existing_roles:
            if role.name == rolename:
                return role
        return None

    def create_user(self, username, password=None, project=None, domain='default'):
        """
        Create a new user
        Invalid groupname : raise keystoneclient.exceptions.NotFound

        project - The tenant/project to assign user when creating (opt)
        """
        account_data = {
            'name': username,
            'password': password,
            'email': '%s@iplantcollaborative.org' % username,
        }
        if project:
            if self.keystone_version() == 3:
                account_data['project'] = project.name
                account_data['domain'] = domain
            elif self.keystone_version() == 2:
                account_data['tenant_id'] = project.id
        return self.keystone.users.create(**account_data)

    ##DELETE##
    def delete_role(self, rolename):
        """
        Retrieve,Delete the user
        Invalid username : raise keystoneclient.exceptions.NotFound
        """
        role = self.get_role(rolename)
        if role:
            role.delete()
        return True

    def delete_project(self, groupname):
        """
        Retrieve and delete the project/group matching groupname
        Returns True on success
        Invalid groupname : raise keystoneclient.exceptions.NotFound
        """
        project = self.get_project(groupname)
        if project:
            project.delete()
        return True

    def delete_all_roles(self, username, projectname):
        project = self.get_project(projectname)
        user = self.get_user(username)
        #roles = self.list_roles(user=user, project=project)
        roles = self.list_roles()
        for role in roles:
            try:
                project.remove_user(user, role)
            except NovaNotFound, NotFound:
                continue


    def delete_project_member(self, groupname, username, adminRole=False):
        """
        Retrieves the project and user object
        Removes user of the admin/member role
        Returns True on success
        Invalid username, groupname, rolename:
            raise keystoneclient.exceptions.NotFound
        """
        project = self.get_project(groupname)
        user = self.get_user(username)
        #FIXME: Hardcoded values
        if adminRole:
            role = self.get_role('admin')
        else:
            role = self.get_role('defaultMemberRole')
        if not project or not user:
            return True
        try:
            project.remove_user(user, role)
            return True
        except NotFound as no_role_for_user:
            logger.debug('Error - %s: User-role combination does not exist' %
                         no_role_for_user)
            return True
        except Exception, e:
            logger.exception(e)
            raise

    def delete_user(self, username):
        """
        Retrieve,Delete the user
        Invalid username : raise keystoneclient.exceptions.NotFound
        """
        user = self.get_user(username)
        if user:
            user.delete()
        return True

    def get_role(self, rolename):
        """
        Retrieve role
        Invalid rolename : raise keystoneclient.exceptions.NotFound
        """
        try:
            return find(self.keystone.roles, name=rolename)
        except NotFound:
            return None

    def get_project_by_id(self, project_id, **kwargs):
        """
        Retrieve project
        Invalid groupname : raise keystoneclient.exceptions.NotFound
        """
        if self.keystone_version() == 2:
            kwargs.pop('domain_id', 'default')  # Remove the kwarg
        try:
            return find(self.keystone_projects(), id=project_id, **kwargs)
        except NotFound:
            return None

    def get_project(self, project_name, **kwargs):
        """
        Retrieve project
        Invalid project_name : raise keystoneclient.exceptions.NotFound
        """
        if self.keystone_version() == 2:
            kwargs.pop('domain_id', 'default')  # Remove the kwarg
        try:
            return find(self.keystone_projects(), name=project_name, **kwargs)
        except NotFound:
            return None

    def get_user(self, username, **kwargs):
        """
        Retrieve user
        Invalid username : raise keystoneclient.exceptions.NotFound
        """
        if self.keystone_version() == 3:
            domain = kwargs.pop('domain', 'default')
            if 'domain_id' not in kwargs:
                kwargs['domain_id'] = domain
        try:
            return self.keystone.users.find(name=username, **kwargs)
        except NotFound:
            return None

    def list_roles(self, **kwargs):
        """
        See keystoneclient for all kwarguments in List roles
        The basics:
        * If passing user=, you must pass project= or domain= for 'pairings'
        * no kwargs == all roles.
        """
        return self.keystone.roles.list(**kwargs)

    def list_projects(self, **kwargs):
        return self.keystone_projects().list(**kwargs)

    def keystone_projects(self):
        if self.keystone_version() == 3:
            return self.keystone.projects
        elif self.keystone_version() == 2:
            return self.keystone.tenants

    def keystone_version(self):
        if self.keystone.version == 'v3':
            return 3
        elif self.keystone.version == 'v2.0':
            return 2

    def list_users(self):
        return self.keystone.users.list()




"""
UserManager:
  Remote Openstack  Admin controls..


DEPRECATION NOTE: THIS IS NOT THE ACCOUNTS/OPENSTACK.PY THAT YOU ARE LOOKING FOR!
YOU WANT THE `accounts/openstack.py` in ATMOSPHERE!

THIS FILE WILL GO AWAY VERY SOON!
"""
from hashlib import sha1
from urlparse import urlparse

from novaclient.v2 import client as nova_client
from novaclient.exceptions import OverLimit

from threepio import logger

from rtwo import settings

from rtwo.drivers.openstack_user import UserManager
from rtwo.drivers.openstack_network import NetworkManager
from rtwo.drivers.common import _connect_to_glance, _connect_to_nova,\
    _connect_to_keystone


class AccountDriver():
    user_manager = None
    network_manager = None

    def __init__(self, *args, **kwargs):
        network_args = {}


    def get_openstack_clients(self, username, password=None, tenant_name=None):

        user_creds = self._get_openstack_credentials(username,
                                                     password,
                                                     tenant_name)
        neutron = self.network_manager.new_connection(**user_creds)
        args = ()
        kwargs = user_creds
        keystone = _connect_to_keystone(*args, **kwargs)
        nova = _connect_to_nova(*args, **kwargs)
        glance = _connect_to_glance(keystone, *args, **kwargs)
        return {
            'glance': glance,
            'keystone': keystone,
            'nova': nova,
            'neutron': neutron,
            'horizon': self._get_horizon_url(keystone.tenant_id)
        }

    def _get_horizon_url(self, tenant_id):
        parsed_url = urlparse("")
        return 'https://%s/horizon/auth/switch/%s/?next=/horizon/project/' %\
            (parsed_url.hostname, tenant_id)

    def create_account(self, username, admin_role=False, max_quota=False):
        """
        Create (And Update 'latest changes') to an account

        """
        finished = False
        # Special case for admin.. Use the Openstack admin identity..
        if username == 'admin':
            ident = self.create_openstack_identity('','','')
            return ident
        #Attempt account creation
        while not finished:
            try:
                password = self.hashpass(username)
                # Retrieve user, or create user & project
                user = self.get_or_create_user(username, password,
                                               True, admin_role)
                logger.debug(user)
                project = self.get_project(username)
                logger.debug(project)
                roles = user.list_roles(project)
                logger.debug(roles)
                if not roles:
                    self.user_manager.add_project_member(username,
                                                         username,
                                                         admin_role)
                self.user_manager.build_security_group(
                    user.name,
                    self.hashpass(user.name),
                    project.name)

                finished = True

            except OverLimit:
                print 'Requests are rate limited. Pausing for one minute.'
                time.sleep(60)  # Wait one minute
        #(user, group) = self.create_usergroup(username)
        return (user, password, project) # was user

    # Useful methods called from above..
    def get_or_create_user(self, username, password=None,
                           usergroup=True, admin=False):
        user = self.get_user(username)
        if user:
            return user
        user = self.create_user(username, password, usergroup, admin)
        return user

    def create_user(self, username, password=None,
                    usergroup=True, admin=False):
        if not password:
            password = self.hashpass(username)
        if usergroup:
            (project, user, role) = self.user_manager.add_usergroup(username,
                                                                    password,
                                                                    True,
                                                                    admin)
        else:
            user = self.user_manager.add_user(username, password)
            project = self.user_manager.get_project(username)
        #TODO: Instead, return user.get_user match, or call it if you have to..
        return user

    def delete_user(self, username, usergroup=True, admin=False):
        project = self.user_manager.get_project(username)
        if project:
            self.network_manager.delete_project_network(username, project.name)
        if usergroup:
            deleted = self.user_manager.delete_usergroup(username)
        else:
            deleted = self.user_manager.delete_user(username)
        return deleted

    def hashpass(self, username):
        return sha1(username).hexdigest()

    def get_project_name_for(self, username):
        """
        This should always map project to user
        For now, they are identical..
        """
        return username

    def get_project(self, project):
        return self.user_manager.get_project(project)

    def list_projects(self):
        return self.user_manager.list_projects()

    def get_user(self, user):
        return self.user_manager.get_user(user)

    def list_users(self):
        return self.user_manager.list_users()

    def list_usergroup_names(self):
        return [user.name for (user, project) in self.list_usergroups()]

    def list_usergroups(self):
        users = self.list_users()
        groups = self.list_projects()
        usergroups = []
        for group in groups:
            for user in users:
                if user.name in group.name:
                    usergroups.append((user, group))
                    break
        return usergroups

    def _get_openstack_credentials(self, username, password=None,
                                   tenant_name=None):
        if not tenant_name:
            tenant_name = self.get_project_name_for(username)
        if not password:
            password = self.hashpass(tenant_name)
        user_creds = {'auth_url': self.user_manager.nova.client.auth_url,
                      'region_name': self.user_manager.nova.client.region_name,
                      'username': username,
                      'password': password,
                      'tenant_name': tenant_name}
        return user_creds



"""
Atmosphere service size.

"""
from abc import ABCMeta

from rtwo.models.provider import AWSProvider, EucaProvider, OSProvider
from threepio import logger


class BaseSize(object):
    __metaclass__ = ABCMeta


class Size(BaseSize):

    provider = None

    sizes = {}

    lc_sizes = None

    def __init__(self, lc_size):
        self._size = lc_size
        self.id = self._size.id
        self.name = self._size.name
        self.price = self._size.price
        self.ram = self._size.ram
        self.disk = self._size.disk
        if hasattr(self._size, 'extra'):
            self.extra = self._size.extra
        else:
            self.extra = {}  # Placeholder Dict
        self.cpu = self.extra.get('cpu',0)
        self.ephemeral = self.extra.get('ephemeral',0)
        self.bandwidth = 0

    @classmethod
    def create_size(cls, provider, lc_size):
        size = provider.sizeCls(lc_size)
        alias = size.id
        cls.sizes[(provider.identifier, alias)] = size
        return size

    @classmethod
    def lookup_size(cls, alias, provider):
        if cls.sizes.get((provider.identifier, alias)):
            return cls.sizes[
                (provider.identifier, alias)
            ]
        else:
            return None

    @classmethod
    def get_size(cls, lc_size, provider):
        alias = lc_size.id
        if cls.sizes.get((provider.identifier, alias)):
            return cls.sizes[
                (provider.identifier, alias)
            ]
        else:
            return cls.create_size(provider, lc_size)

    @classmethod
    def get_sizes(cls, provider, lc_list_sizes_method):
        identifier = provider.identifier
        cached_sizes = cls.sizes.get(identifier)
        if not cached_sizes or not cls.lc_sizes:
            cls.lc_sizes = lc_list_sizes_method()
            logger.debug("Caching %s sizes for identifier:%s" %
                         (len(cls.lc_sizes), identifier))
        return sorted(
            [cls.get_size(size, provider) for size in cls.lc_sizes],
            key=lambda s: (s.cpu, s.ram))

    def reset(self):
        Size.reset()
        self._size = None
        self.lc_sizes = None
        self.sizes = {}

    @classmethod
    def reset(cls):
        cls.lc_sizes = None
        cls.sizes = {}

    def __unicode__(self):
        return str(self)

    def __str__(self):
        return reduce(
            lambda x, y: x+y,
            map(unicode, [self.__class__, " ", self.json()]))

    def __repr__(self):
        return str(self)

    def json(self):
        return {
            'id': self.name,
            'provider': self.provider.identifier,
            'alias': self.id,
            'name': self.name,
            'cpu': self.cpu,
            'ram': self.ram,
            'root': self.disk,
            'disk': self.ephemeral,
            'bandwidth': self.bandwidth,
            'price': self.price}

class MockSize(Size):
    def __init__(self, size_id, provider):
        self.provider = provider
        self._size = None
        self.name = "Unknown Size %s" % size_id
        self.alias = size_id
        self.id = size_id
        self.price = None
        self.ram = 0
        self.disk = 0
        self.extra = {}  # Placeholder Dict
        self.cpu = 0
        self.ephemeral = 0

    def json(self):
        return {
            'id': self.id,
            'provider': self.provider.identifier,
            'alias': self.id,
            'name': 'MockSize %s' % self.id,
            'cpu': self.cpu,
            'ram': '',
            'root': '',
            'disk': '',
            'bandwidth': '',
            'price': ''}

class EucaSize(Size):

    provider = EucaProvider


class AWSSize(Size):

    provider = AWSProvider


class OSSize(Size):

    provider = OSProvider



"""
Atmosphere service identity.

"""
from abc import ABCMeta, abstractmethod

from threepio import logger

from rtwo.exceptions import MissingArgsException
from rtwo.models.provider import AWSProvider, EucaProvider, OSProvider


class BaseIdentity(object):
    __metaclass__ = ABCMeta

    provider = None

    groups = []
    providers = []
    machines = []
    instances = []

    credentials = {}

    @abstractmethod
    def __init__(self, provider, user, key, secret):
        raise NotImplemented


class Identity(BaseIdentity):

    def __init__(self, provider, key=None, secret=None, user=None, **kwargs):
        if issubclass(type(provider), self.provider):
            self.providers.append(provider)
        else:
            logger.warn("Provider doesn't match (%s != %s)." %
                        (provider, self.provider))
        self.user = user
        self.credentials = {}
        self.credentials.update(kwargs)
        self.credentials.update({'key': key, 'secret': secret})

    def get_username(self):
        return self.credentials['key']
    def get_groupname(self):
        return self.credentials['key']
    def __repr__(self):
        return '%s (%s) Credentials: %s' % (self.__class__,
                                            self.credentials.get('key',''),
                                            self.credentials.keys())



class AWSIdentity(Identity):

    provider = AWSProvider


class EucaIdentity(Identity):

    provider = EucaProvider


class OSIdentity(Identity):

    provider = OSProvider
